from pyppl import Box
from bioprocs.utils import shell

params = {{args.params | repr}}
params.i = {{i.infile | quote}}
params._stdout = {{o.outfile | quote}}

shell.TOOLS.bedtools = {{args.bedtools | quote}}
shell.Shell(subcmd = True, dash = '-', equal = '=').bedtools.merge(**params).run()




from sys import stderr
from pyppl import Box
from bioprocs.utils import shell2 as shell

params = {{args.params | repr}}

params['a']           = {{i.afile | quote}}
params['b']           = {{i.bfile | quote}}
params['wa']          = params.get('wa', True)
params['wb']          = params.get('wb', True)
params['nonamecheck'] = params.get('nonamecheck', True)
params['_out']        = {{o.outfile | quote}}
params['_stderr']     = stderr

shell.load_config(bedtools = {{args.bedtools | quote}})
shell.bedtools.intersect(**params)



from pyppl import Box
from bioprocs.utils.shell import grep, zcat

infile  = {{ i.infile | quote}}
outfile = {{ o.outfile | quote}}
params  = {{ args.params | repr}}
keyword = {{ args.keyword | repr}}
if not keyword:
	raise ValueError('A keyword (args.keyword) is required.')

params._stdout = outfile
if infile.endswith('.gz'):
	zcat(infile).pipe().grep(_ = keyword, **params)
else:
	grep(_ = [keyword, infile], **params)



from os import path
from bioprocs.utils import FileConn
from bioprocs.utils.tsvio2 import TsvReader

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
ref     = {{args.ref | quote}}
reffai  = ref + '.fai'
refdict = ref[:-3] + '.dict'

if not path.isfile(refdict) and not path.isfile(reffai):
	raise OSError('A dict file or fai file not exists for the reference file.')

contigs = []
if path.isfile(refdict):
	reader = TsvReader(refdict, skip = 1, cnames = False)
	for r in reader:
		contigs.append(r[1][3:])

elif path.isfile(reffai):
	reader = TsvReader(reffai, cnames = False)
	for r in reader:
		contigs.append(r[0])

with FileConn(infile, 'rt') as f, open(outfile, 'w') as fout:
	for line in f:
		# also remove contigs from header
		if line.startswith('##contig=<ID='):
			contig = line[13:].split(',')[0].split('>')[0]
			if contig in contigs:
				fout.write(line)
		elif line.startswith('#'):
			fout.write(line)
		else:
			config, _ = line.split('\t', 1)
			if config in contigs:
				fout.write(line)



from pyppl import Box
from os import path

def dsnparse(dsn):
	ret = Box()
	scheme, args = dsn.split(':', 1)
	ret.scheme = scheme
	args = args.split(';')
	for arg in args:
		arg = arg.strip()
		if not arg: continue
		argname, argval = arg.split('=')
		ret[argname] = argval
	if not hasattr(ret, 'after'):
		ret.after = ''
	return ret

def schemaparse(schemafile, type = 'schema', scheme = 'sqlite', delimit = '\t'):
	name, fields = None, []
	name = path.basename(schemafile).split('.')[0]
	if type == 'schema':
		with open(schemafile) as f:
			for line in f:
				line = line.strip(' \n')
				if not line: continue
				if line.startswith('#'):
					line = line.strip()
					if line.startswith('Name:'):
						name = line[5:].strip()
					else:
						continue
				parts = line.split(delimit)
				if len(parts) == 2: parts.append('')
				fields.append(parts)
	else: # type == 'data'
		if scheme not in ['sqlite', 'sqlite3']:
			raise NotImplementedError('Only support sqlite3 currently.')
		with open(schemafile) as f:
			header = f.readline().strip('# \n').split(delimit)
		for head in header:
			fields.append([head, 'TEXT', ''])
	return name, fields



from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

cnsfile  = {{i.cnsfile | quote}}
cnnfile  = {{i.cnnfile | quote}}
outfile  = {{o.outfile | quote}}
nthread  = {{args.nthread | repr}}
params   = {{args.params | repr}}
cnvkit   = {{args.cnvkit | quote}}

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = path.dirname(outfile)
))

# region cnvkit export example
# cnvkit.py export theta Sample_T.cns reference.cnn -v Sample_Paired.vcf
# cnvkit.py export theta Sample_Tumor.cns Sample_Normal.cnr -o Sample.theta2.interval_count
# cnvkit.py export theta Sample_Tumor.cns -o Sample.theta2.interval_count
# endregion

# region cnvkit export theta usage
# usage: cnvkit.py export theta [-h] [-r REFERENCE] [-o OUTPUT] [-v VCF]
#                               [-i SAMPLE_ID] [-n NORMAL_ID]
#                               [-m MIN_VARIANT_DEPTH] [-z [ALT_FREQ]]
#                               tumor_segment

# positional arguments:
#   tumor_segment         Tumor-sample segmentation file from CNVkit (.cns).

# optional arguments:
#   -h, --help            show this help message and exit
#   -r REFERENCE, --reference REFERENCE
#                         Reference copy number profile (.cnn), or normal-sample
#                         bin-level log2 copy ratios (.cnr). Use if the
#                         tumor_segment input file does not contain a "weight"
#                         column.
#   -o OUTPUT, --output OUTPUT
#                         Output file name.

# To also output tables of SNP b-allele frequencies for THetA2:
#   -v VCF, --vcf VCF     VCF file containing SNVs observed in both the tumor
#                         and normal samples. Tumor sample ID should match the
#                         `tumor_segment` filename or be specified with -i
#                         /--sample-id.
#   -i SAMPLE_ID, --sample-id SAMPLE_ID
#                         Specify the name of the tumor sample in the VCF (given
#                         with -v/--vcf). [Default: taken the tumor_segment file
#                         name]
#   -n NORMAL_ID, --normal-id NORMAL_ID
#                         Corresponding normal sample ID in the input VCF.
#   -m MIN_VARIANT_DEPTH, --min-variant-depth MIN_VARIANT_DEPTH
#                         Minimum read depth for a SNP in the VCF to be counted.
#                         [Default: 20]
#   -z [ALT_FREQ], --zygosity-freq [ALT_FREQ]
#                         Ignore VCF's genotypes (GT field) and instead infer
#                         zygosity from allele frequencies. [Default if used
#                         without a number: 0.25]
# endregion

params.o   = path.basename(outfile)
if cnnfile:
	params.r = cnnfile

shell.fg.cnvkit.export('theta', cnsfile, **params)




from pyppl import Box
from bioprocs.utils import shell2 as shell

infile   = {{i.infile | repr}}
outfile  = {{o.outfile | repr}}
umfile   = {{o.umfile | repr}}
params   = {{args.params | str}}
chain    = {{args.lochain | repr}}
liftover = {{args.liftover | repr}}

shell.load_config(liftover = liftover)

shell.fg.liftover(infile, chain, outfile, umfile, **params)



from pyppl import Box
from bioprocs.utils import shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
params   = {{args.params | repr}}
bedtools = {{args.bedtools | quote}}

bedtools = shell.Shell({'bedtools': bedtools}, subcmd = True, equal = ' ', dash = '-').bedtools

params.ibam    = infile
params._stdout = outfile
bedtools.genomecov(**params).run()


from os import path
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
samfile = {{i.samfile | quote}}
outfile = {{o.outfile | quote}}
samples = {{args.samples | repr}}

if samfile and path.isfile(samfile):
	with open(samfile) as f:
		samples = f.read().strip().splitlines()
elif samfile:
	samples = samfile.split(',')
elif not isinstance(samples, list):
	samples = samples.split(',')

reader = TsvReader(infile)
writer = TsvWriter(outfile)
writer.cnames = reader.cnames
writer.writeHead()

for r in reader:
	if r.Tumor_Sample_Barcode in samples:
		writer.write(r)

reader.close()
writer.close()



"""ChIP-seq data analysis"""
from pyppl import Proc
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pPeakToRegPotential():
	"""
	@name:
		pPeakToRegPotential
	@description:
		Convert peaks to regulatory potential score for each gene
		The formula is:
		```
						-(0.5 + 4*di/d0)
		PC = sum (pi * e                  )
		```
		Ref: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4489297/
	@input:
		`peakfile:file`: The BED/peak file for peaks
		`genefile:file`: The BED file for gene coordinates
	@output:
		`outfile:file`: The regulatory potential file for each gene
	@args:
		`signal`: `pi` in the formula. Boolean value, whether use the peak intensity signale or not, default: `True`,
		`genefmt`: The format for `genefile`, default: `ucsc+gz`. It could be:
			- ucsc or ucsc+gz: typically, you can download from http://hgdownload.cse.ucsc.edu/goldenPath/hg38/database/refGene.txt.gz
			- bed or bed+gz: [format](https://genome.ucsc.edu/FAQ/FAQformat#format1), 4th column required as gene identity.
		`peakfmt`: The format for `peakfile`, default: `peak`. It could be:
			- peak or peak+gz: (either [narrowPeak](https://genome.ucsc.edu/FAQ/FAQformat.html#format12) or [broadPeak](https://genome.ucsc.edu/FAQ/FAQformat.html#format13), the 7th column will be used as intensity
			- bed or bed+gz: [format](https://genome.ucsc.edu/FAQ/FAQformat#format1), 5th column will be used as intensity.
		`window`: `2 * d0` in the formula. The window where the peaks fall in will be consided, default: `100000`.
			```
					|--------- window ----------|
					|---- d0 -----|
					|--- 50K --- TSS --- 50K ---|
						^ (peak center)
						|-- di --|
			```
	"""
	pPeakToRegPotential              = Proc (desc = 'Convert peaks to regulatory potential score for each gene.')
	pPeakToRegPotential.input        = "peakfile:file, genefile:file"
	pPeakToRegPotential.output       = "outfile:file:{{peakfile | fn}}.rp.txt"
	pPeakToRegPotential.args.signal  = True
	pPeakToRegPotential.args.genefmt = 'ucsc+gz',
	pPeakToRegPotential.args.peakfmt = 'peak',
	pPeakToRegPotential.args.window  = 100000
	pPeakToRegPotential.lang         = params.python.value
	pPeakToRegPotential.script       = "file:scripts/chipseq/pPeakToRegPotential.py"
	return pPeakToRegPotential







from pyppl import Box
from bioprocs.utils.tsvio import TsvReader, TsvWriter, TsvRecord

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts}}
outopts = {{args.outopts}}
match   = {{args.match}}
do      = {{args.do}}

reader = TsvReader(infile, **inopts)
writer = TsvWriter(outfile, **outopts)

if not writer.meta:
	writer.meta = reader.meta

if outopts.head:
	writer.writeHead(
		delimit   = outopts.headDelimit,
		prefix    = outopts.headPrefix,
		transform = outopts.headTransform
	)

prev     = None
repeated = []
for r in reader:
	m = r[0] if not callable(match) else match(r)
	if prev is None:
		prev = m
		repeated.append(r)
	elif prev == m:
		repeated.append(r)
	else:
		prev = m
		writer.write(do(repeated))
		repeated = [r]

if repeated:
	writer.write(do(repeated))

	


from sys import stderr
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs

fq   = {{i.fq | quote}}
params = {{args.params}}
try:
	{% if args.tool == 'fastqc' %}
	params['o'] = {{o.outdir | quote}}
	cmd = '{{args.fastqc}} %s "{{i.fq}}"' % cmdargs(params)
	runcmd(cmd)

	{% else %}
	raise Exception('Tool {{args.tool}} %s not supported.')

	{% endif %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise


from sys import stderr
from shutil import move

from pyppl import Box
from bioprocs.utils import runcmd, mem2, cmdargs

params = {{args.params}}
try:
{% case args.tool %}
	{% when 'trimmomatic' %}
	mem    = mem2 ({{args.mem | quote}}, "java")
	minlen = str({{args.minlen}} * 2)
	adfile = "{{job.outdir}}/adapters.fa"
	with open (adfile, "w") as ad:
		ad.write (">TruSeq3_IndexedAdapter\n")
		ad.write ({{args.adapter | quote}} + "\n")

	params['threads'] = {{args.nthread}}
	cmd    = '{{args.trimmomatic}} %s SE %s "{{in.fq}}" "{{out.outfq}}" ILLUMINACLIP:%s:2:30:10 LEADING:{{args.cut5}} TRAILING:{{args.cut3}} SLIDINGWINDOW:4:{{args.minq}} MINLEN:%s' % (mem, cmdargs(params, dash = '-', equal = ' '), adfile, minlen)
	runcmd (cmd)

	{% when 'cutadapt' %}
	params['a'] = {{args.adapter | quote}}
	params['u'] = "{{args.cut5}}"
	params['u'] = "-{{args.cut3}}"
	params['m'] = {{args.minlen}}
	params['q'] = "{{args.minq}},{{args.minq}}"
	params['o'] = {{out.outfq | quote}}
	cmd = '{{args.cutadapt}} %s "{{in.fq}}"' % cmdargs(params, dash = '-', equal = ' ')
	runcmd (cmd)

	{% when 'skewer' %}
	params['m'] = 'any'
	params['t'] = {{args.nthread}}
	params['x'] = {{args.adapter | quote}}
	params['Q'] = {{args.minq}}
	params['l'] = {{args.minlen}}
	params['z'] = {{args.gz}}
	params['o'] = "{{job.outdir}}/tmp"
	cmd = '{{args.skewer}} %s "{{in.fq}}"' % cmdargs(params, dash = '-', equal = ' ')
	runcmd (cmd)
	outfq = "{{job.outdir}}/tmp-trimmed.fastq"
	{% if args.gz %}
	outfq += ".gz"
	{% endif %}
	move (outfq, "{{out.outfq}}")

	{% else %}
	raise Exception ('Tool {{args.tool}} not supported')
{% endcase %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise



import math, gzip
peakfile = "{{peakfile}}"
genefile = "{{genefile}}"
arg_inst = {{args.signal | repr}}
arg_gf   = "{{args.genefmt}}"
arg_pf   = "{{args.peakfmt}}"
arg_wd   = int({{args.window | repr}})
d0       = arg_wd / 2

assert (isinstance(arg_inst, bool))
assert (arg_gf in ['ucsc', 'bed', 'ucsc+gz', 'bed+gz'])
assert (arg_pf in ['peak', 'bed', 'peak+gz', 'bed+gz'])

open_gf = open_pf = open
if arg_gf.endswith ('+gz'):
	arg_gf  = arg_gf[:-3]
	open_gf = gzip.open
if arg_pf.endswith ('+gz'):
	arg_pf  = arg_pf[:-3]
	open_pf = gzip.open

# read genes
genes = {}
if arg_gf == 'bed':
	with open_gf (genefile) as f:
		for line in f:
			line    = line.strip()
			if not line or line.startswith('track') or line.startswith('#'): continue
			items   = line.split("\t")
			chr     = items[0]
			start   = int(items[1])
			end     = int(items[2])
			gene    = items[3]
			strand  = '-' if len(items)>5 and items[5] == '-' else '+'
			tss     = start if strand == '+' else end
			rstart  = tss - d0
			rend    = tss + d0
			genes[gene] = [chr, start, end, tss, rstart, rend]
else:
	with open_gf (genefile) as f:
		for line in f:
			line    = line.strip()
			if not line or line.startswith('track') or line.startswith('#'): continue
			items   = line.split("\t")
			chr     = items[2]
			start   = int(items[4])
			end     = int(items[5])
			gene    = items[12]
			strand  = items[3]
			tss     = start if strand == '+' else end
			rstart  = tss - d0
			rend    = tss + d0
			genes[gene] = [chr, start, end, tss, rstart, rend]

# read peaks
peaks = {}
if arg_pf == 'peak':
	with open_pf (peakfile) as f:
		for line in f:
			line = line.strip()
			if not line or line.startswith('track') or line.startswith('#'): coninue
			items   = line.split("\t")
			chr     = items[0]
			start   = int(items[1])
			end     = int(items[2])
			signal  = float(items[6])
			if peaks.has_key(chr):
				peaks[chr].append ([start, end, (start+end) / 2, signal])
			else:
				peaks[chr] = [[start, end, (start+end) / 2, signal]]
else:
	with open_pf (peakfile) as f:
		for line in f:
			line = line.strip()
			if not line or line.startswith('track') or line.startswith('#'): coninue
			items   = line.split("\t")
			chr     = items[0]
			start   = int(items[1])
			end     = int(items[2])
			signal  = float(items[4])
			if peaks.has_key(chr):
				peaks[chr].append ([start, end, (start+end) / 2, signal])
			else:
				peaks[chr] = [[start, end, (start+end) / 2, signal]]

for key, val in peaks.iteritems():
	peaks[key] = sorted (val, cmp = lambda x, y: x[0] - y[0])

rp = {}
for gene, ginfo in genes.iteritems():
	(gchr, gstart, gend, gtss, grstart, grend) = ginfo
	rp[gene] = 0
	if not peaks.has_key(gchr): continue
	for pinfo in peaks[gchr]:
		(pstart, pend, pcenter, psignal) = pinfo
		if pcenter < grstart: continue
		if pcenter > grend: break
		score  = psignal if arg_inst else 1
		score *= math.exp (-(.5 + 4*abs(pcenter - tss)/d0))
		rp[gene] += score

with open ("{{outfile}}", 'w') as f:
	for key in sorted (rp, key=rp.get, reverse = True):
		f.write ("%s\t%.3f\n" % (key, rp[key]))


from pyppl import Box
from bioprocs.utils import shell

l        = {{i.l | repr}}
n        = {{i.n | repr}}
seed     = {{args.seed | repr}}
bedtools = {{args.bedtools | quote}}
gsize    = {{args.gsize | quote}}
outfile  = {{o.outfile | quote}}

shell.TOOLS.bedtools = bedtools
shell.Shell(subcmd = True, dash = '-', equal = '=').bedtools.random(l = l, n = n, seed = seed, g = gsize, _stdout = outfile)




from bioprocs.utils.sampleinfo import SampleInfo

infile    = {{i.sifile | quote}}
outfile   = {{o.outfile | quote}}
saminfo   = SampleInfo(infile)
groups    = saminfo.select(get = 'Group')
unigroups = list(set(groups))

with open(outfile, "w") as f:
	f.write("%s\t%s\t1\n" % (saminfo.nrow, len(unigroups)))
	f.write("# %s\n" % (' '.join(unigroups)))
	f.write(' '.join(groups) + '\n')



from os import makedirs, path
from pyppl import Box
from bioprocs.utils import logger, mem2, shell2 as shell

# bam2fastq will create {i.infile}.tmp, use file in indir in case of permission issue
infile  = {{i.infile | quote}}
fqfile1 = {{o.fqfile1 | quote}}
fqfile2 = {{o.fqfile2 | quote}}
tool    = {{args.tool | quote}}
params  = {{args.params | repr}}
mem     = {{ args.mem | quote }}
gz      = {{args.gz | bool}}
tmpdir  = {{args.tmpdir | quote}}

biobambam = {{args.biobambam | quote}}
bedtools  = {{args.bedtools | quote}}
samtools  = {{args.samtools | quote}}
picard    = {{args.picard | quote}}

shell.load_config(
	biobambam = biobambam,
	bedtools  = bedtools,
	samtools  = samtools,
	picard    = picard
)

if gz:
	fqfile1 = fqfile1[:-3]
	fqfile2 = fqfile2[:-3]

tmpdir = path.join(tmpdir, "tmp.{{proc.id}}.{{proc.tag}}.{{proc.suffix}}.{{job.index}}")
if not path.exists(tmpdir):
	makedirs(tmpdir)

params._out_ = {{job.outfile | quote}}
params._err_ = {{job.errfile | quote}}

def gzip_fqs():
	if gz:
		shell.gzip(fqfile1)
		shell.gzip(fqfile2)

def run_biobambam():
	params.gz       = 0
	params.F        = fqfile1
	params.F2       = fqfile2
	params.T        = path.join(tmpdir, infile + '.tmp')
	params.filename = infile
	if infile.endswith('.sam'):
		params.inputformat = 'sam'
	shell.biobambam(**params)
	gzip_fqs()

def run_bedtools():
	params.i = infile
	params.fq = fqfile1
	params.fq2 = fqfile2
	shell.bedtools.bamtofastq(**params)
	gzip_fqs()

def run_samtools():
	params.t = True
	params['1'] = fqfile1
	params['2'] = fqfile2
	params._ = infile
	shell.samtools.fastq(**params)
	gzip_fqs()

def run_picard():
	mem = mem2({{ args.mem | quote }}, 'Java')
	params[mem]                = True
	params['-Djava.io.tmpdir'] = tmpdir
	params.TMP_DIR = tmpdir
	params.I = infile
	params.F = fqfile1
	params.F2 = fqfile2
	shell.picard(**params)
	gzip_fqs()

tools = dict(
	biobambam = run_biobambam,
	bedtools = run_bedtools,
	samtools = run_samtools,
	picard = run_picard,
)

try:
	tools[tool]()
except KeyError:
	raise ValueError('Tool "%r" not supported.' % tool)
finally:
	shell.rm_rf(tmpdir)



from pyppl import Box
from bioprocs.utils.gene import genenorm
from bioprocs.utils.tsvio import TsvReader, TsvWriter

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
notfound = {{args.notfound | quote}}
inopts   = {{args.inopts}}
outopts  = {{args.outopts}}
genecol  = {{args.genecol | repr}}
fromipi  = {{args.fromipi | bool}}
ipidb    = {{args.ipidb | quote}}

if genecol is None:
	genecol = 0

# get dicts 
ipireader = TsvReader(ipidb, ftype = 'nometa')
dicts = {}
for r in ipireader:
	gene = r[6]
	ipis = r[9].split(';')
	if not gene: continue
	if fromipi:
		dicts.update({ipi:gene for ipi in ipis if ipi and gene})
	else:
		dicts[gene] = '|'.join(ipi for ipi in ipis if ipi)

reader = TsvReader(infile, **inopts)
writer = TsvWriter(outfile, **outopts)
writer.meta = reader.meta
if outopts.query:
	writer.meta.add('_QUERY')

if outopts.head:
	writer.writeHead(
		delimit   = outopts.headDelimit,
		prefix    = outopts.headPrefix,
		transform = outopts.headTransform)

for r in reader:
	query = r[genecol]
	if not query in dicts:
		if notfound == 'ignore':
			ret = query
		elif notfound == 'skip':
			continue
		elif notfound == 'error':
			raise ValueError('{} not found!'.format(query))
	else:
		ret = dicts[query]
	r[genecol] = ret
	if outopts.query:
		if isinstance(r, list):
			r.append(query)
		else:
			r._QUERY = query
	writer.write(r)
	



from bioprocs.utils.meme import MemeReader, MemeWriter

infiles = {{ i.infiles | repr}}
outfile = {{ o.outfile | quote}}

reader0 = MemeReader(infiles[0])
writer  = MemeWriter(outfile)
writer.meta = reader0.meta
writer.writeMeta()
reader0.close()

for infile in infiles:
	reader = MemeReader(infile)
	for r in reader:
		writer.write(r)
	reader.close()

writer.close()




import gzip
from os import path, symlink, readlink, remove
from bioprocs.utils import runcmd, cmdargs
from subprocess import check_output
from bioprocs.utils.parallel import Parallel

infile1  = {{i.infile1 | quote}}
infile2  = {{i.infile2 | quote}}
outfile  = {{o.outfile | quote}}
rmany    = {{args.any | repr}}
tool     = {{args.tool | quote}}
header   = {{args.header | repr}}
bedtools = {{args.bedtools | quote}}
tabix    = {{args.tabix | quote}}
bychrom  = {{args.bychrom | repr}}
nthread  = int({{args.nthread | repr}})
outdir   = {{job.outdir | quote}}

if tool == 'bedtools':
	if header:
		# write the header to outfile
		openfunc = gzip.open if infile1.endswith('.gz') else open
		with openfunc(infile1) as fin, open(outfile, 'w') as fout:
			for line in fin:
				if not line.startswith('#'):
					break
				fout.write(line)

	# go directly
	if not bychrom:
		cmd = '{bedtools} subtract {params} >> {outfile}'
		params = {
			'a': infile1,
			'b': infile2
		}
		if rmany:
			params['A'] = True
		
		params = cmdargs(params, dash = '-', equal = ' ')
		runcmd(cmd.format(
			bedtools = bedtools,
			params   = params,
			outfile  = outfile
		))
	else:
		# check if infile1 is tabix indexed
		def tabindex(vcf, outdir):
			if not vcf.endswith('.gz'):
				gzfile = path.join(outdir, path.basename(vcf) + '.gz')
				bgzip_cmd = 'bgzip {!r} -c > {!r}'.format(
					vcf, 
					gzfile
				)
				runcmd(bgzip_cmd)
				runcmd('{} {!r}'.format(tabix, gzfile))
			else:
				gzfile = path.join(outdir, path.basename(vcf))
				# it is gzipped, try to find the index file (.tbi)
				idxfile = path.join(outdir, path.basename(vcf) + '.tbi')
				symlink(vcf, gzfile)
				while True:
					try:
						link = readlink(vcf)
						tbifile = link + '.tbi'
						if path.isfile(tbifile):
							symlink(tbifile, gzfile + '.tbi')
							break
						vcf = link
					except OSError:
						break
				
				if not path.isfile(idxfile):
					index_cmd = '{} {!r}'.format(tabix, gzfile)
					runcmd(index_cmd)

			return gzfile

		def runChrom(file1, file2, chrom):
			outfile1_list = list(file1[:-3].rpartition('.'))
			outfile1_list.insert(-2, '-' + chrom)
			outfile1 = ''.join(outfile1_list)
			outfile2 = list(file2[:-3].rpartition('.'))
			outfile2.insert(-2, '-' + chrom)
			outfile2 = ''.join(outfile2)
			outfile1_list.insert(-2, '.subtracted')
			outfile  = ''.join(outfile1_list)
			vfcmd    = '{} -h {!r} {!r} > {!r}'
			runcmd(vfcmd.format(tabix, file1, chrom, outfile1))
			runcmd(vfcmd.format(tabix, file2, chrom, outfile2))

			cmd = '{bedtools} subtract {params} > {outfile!r}'
			params = {
				'a': outfile1,
				'b': outfile2
			}
			if rmany:
				params['A'] = True
			
			params = cmdargs(params, dash = '-', equal = ' ')
			runcmd(cmd.format(
				bedtools = bedtools,
				params   = params,
				outfile  = outfile
			))
			remove(outfile1)
			remove(outfile2)
			return outfile
		
		infile1 = tabindex(infile1, outdir)
		infile2 = tabindex(infile2, outdir)
		chroms  = [chr.strip() for chr in check_output([tabix, '-l', infile1]).splitlines()]

		if nthread > 1:
			p = Parallel(nthread, raiseExc = True)
			outfiles = p.run(runChrom, [(infile1, infile2, chrom) for chrom in chroms])
			# make sure it's in the right order
			outfiles = sorted(outfiles, key = lambda x: chroms.index(x.split('.')[-2]))
		else:
			outfiles = []
			for chrom in chroms:
				outfiles.append(runChrom(infile1, infile2, chrom))

		with open(outfile, 'a+') as fout:
			for of in outfiles:
				with open(of) as f:
					fout.write(f.read())
		
elif tool == 'pyvcf':
	import vcf
	v1  = vcf.Reader(filename = infile1)
	v2  = vcf.Reader(filename = infile2)
	out = vcf.Writer(open(outfile, 'w'), v1)
	r1  = r2 = None
	while True:
		try:
			r1 = r1 or v1.next()
			r2 = r2 or v2.next()
			if r1.CHROM < r2.CHROM:
				out.write_record(r1)
				r1 = None
				continue
			if r1.CHROM > r2.CHROM:
				r2 = None
				continue
			if r1.POS < r2.POS:
				out.write_record(r1)
				r1 = None
				continue
			if r1.POS > r2.POS:
				r2 = None
				continue
			r1 = None
		except StopIteration:
			break
		except:
			r1 = r2 = None
			continue
else:
	import gzip
	with open(infile1) if not infile1.endswith('.gz') else gzip.open(infile1) as f1, \
		 open(infile2) if not infile2.endswith('.gz') else gzip.open(infile2) as f2, \
		 open(outfile, 'w') as fout:
		line1 = line2 = None
		for line in f1:
			if line.startswith('#'):
				fout.write(line)
			else:
				line1 = line.split('\t')
				break
		
		while True:
			try:
				line1 = line1 or f1.next().split('\t')
				line2 = line2 or f2.next().split('\t')
				if line1[0] < line2[0]:
					fout.write('\t'.join(line1))
					line1 = None
					continue
				if line1[0] > line2[0]:
					line2 = None
					continue
				if int(line1[1]) < int(line2[1]):
					fout.write('\t'.join(line1))
					line1 = None
					continue
				if int(line1[1]) > int(line2[1]):
					line2 = None
					continue
				line1 = None
			except StopIteration:
				break
			except:
				line1 = line2 = None
				continue
			





#!/usr/bin/env python
"""Using PyPPL to distribute and run commands."""
import sys
from os import path
from pyppl import PyPPL, Proc, Box
from bioprocs import params
from tempfile import gettempdir

params._prefix     = '-'
params._hbald      = False
params.cmds.desc   = 'The cmd list. If not provided, STDIN will be used.'
params.runner      = 'local'
params.runner.desc = 'The runner.'
params.intype      = 'stdin' # or file, stdin, or cmds (pass cmds directly)
params.intype.desc = 'Type of option `cmds`, cmds or file?'
params.cache       = False
params.cache.desc  = 'Cache the jobs or not?'
params.ppldir      = path.join(gettempdir(), 'bioprocs.workdir')
params.ppldir.desc = 'The ppldir to save the pipeline data.'
params.forks       = 1
params.forks.desc  = 'How many jobs to run simultaneously.'

def main():
	opts = params._parse(dict_wrapper = Box)

	if not isinstance(opts.cmds, list):
		if opts.intype == 'cmds':
			opts.cmds = opts.cmds.splitlines()
		elif opts.intype == 'file':
			with open(opts.cmds) as f:
				opts.cmds = f.read().splitlines()
		else:
			opts.cmds = []
			for line in sys.stdin:
				opts.cmds.append(line.strip())

	pCmdRunner        = Proc(desc = 'Using PyPPL to distribute and run commands.')
	pCmdRunner.runner = opts.get('runner', 'local')
	pCmdRunner.cache  = opts.cache
	pCmdRunner.errhow = 'halt'
	pCmdRunner.forks  = int(opts.get('forks', 1))
	pCmdRunner.input  = {'cmd': opts.cmds}
	pCmdRunner.script = '{{i.cmd}}'

	config = {'_log': {'file': None}, 'default': {'ppldir': opts.ppldir}}
	PyPPL(config).start(pCmdRunner).run()

if __name__ == "__main__":
	main()



from pyppl import Box
from bioprocs.utils import shell

infile    = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
intype    = {{args.intype | quote}}
params   = {{args.params | repr}}
bedtools = {{args.bedtools | quote}}

bedtools = shell.Shell({'bedtools': bedtools}, subcmd = True, equal = ' ', dash = '-').bedtools

if intype == 'bed':
	params.b = infile
else:
	params.g = infile

params._       = intfile
params._stdout = outfile
bedtools.makewindows(**params).run()



from splinter import Browser
import json, os, sys, time


browser = Browser('phantomjs')

try:
	data    = {{i.data}}

	sys.stderr.write ('Visiting {{i.url | quote}} ...\n')
	browser.visit({{i.url | quote}})

	sys.stderr.write ("Filling parameters ...\n")
	for key, val in data.iteritems():
		print key
		if key.startswith('/'):
			element = browser.find_by_xpath(key)
			if not element:
				raise KeyError('Cannot find element: %s' % key)
			element.fill(val)
		else:
			browser.fill(key, val)

	i = 0
	browser.find_by_xpath({{i.submit | quote}}).first.click()

	while {{i.next | lambda x: bool(x)}} and \
		browser.is_element_present_by_xpath ({{i.next | quote}}):
		sys.stderr.write ("Fetching and saving page %s ...\n" % (i+1))
		with open ("{{o.outdir}}/out-%s.html" % i, 'w') as f:
			f.write (browser.html.encode('utf-8'))
		i += 1
		time.sleep({{args.interval}})
		browser.find_by_xpath({{i.next | quote}}).first.click()

	sys.stderr.write ("Fetching and saving page %s ...\n" % (i+1))
	with open ("{{o.outdir}}/out-%s.html" % i, 'w') as f:
		f.write (browser.html.encode('utf-8'))
	sys.stderr.write ("Qutting browser ...\n")
except Exception:
	raise
finally:
	browser.quit()



from collections import Counter
from pysam import VariantFile as Vcf
from bioprocs.utils.tsvio2 import TsvJoin

infile   = {{i.infile     | quote}}
outfile  = {{o.outfile    | quote}}
useid    = {{args.useid   | repr}}
na       = {{args.na      | quote}}
mingt    = {{args.mingt   | repr}}
novel    = {{args.novel   | quote}}
dbsnp    = {{args.dbsnp   | repr}}
bialt    = {{args.bialt   | repr}}
chrorder = {{args.chrorder.split(',') | repr}}
samname  = {{args.samname }}

chrorder = dict(zip(chrorder, list(range(len(chrorder)))))
if dbsnp:
	tsvjoin = TsvJoin(dbsnp, infile, cnames = False)
vcf = Vcf(infile)
samples = list(vcf.header.samples)
if callable(samname):
	samples = [samname(sam) for sam in samples]
sc = Counter(samples)
for s, c in sc.items():
	if c > 1:
		raise ValueError('Duplicated sample: {!r}'.format(s))

with open(outfile, 'w') as fout:
	fout.write("\t".join(samples) + "\n")
if mingt <= 1:
	mingt = len(samples) * mingt

def formatGT(gt):
	gt0 = gt[0]
	gt1 = gt[-1]
	try:
		gt0 = int(gt0)
		gt1 = int(gt1)
	except (ValueError, TypeError):
		return na
	gt = gt0 + gt1
	if bialt and gt > 2:
		return na
	return str(gt)

def formatRow(row, rs = novel):
	if bialt and row[3] not in list('AGTCagtc'):
		return None
	row[4] = row[4].rstrip(' ,')
	if bialt:
		row[4] = row[4].split(',')[0]
		if row[4] not in list('AGTCagtc'):
			return None
	gts = [formatGT(gt) for gt in row[9:]]
	if mingt > 0 and sum(1 for gt in gts if gt!=na) < mingt:
		return None
	if not useid or not row[2] or row[2] == '.':
		row[2] = rs
	if not rs: return None
	return "_".join(row[:5]) + "\t" + "\t".join(gts) + "\n"

if not dbsnp:
	reader = open(infile)
	writer = open(outfile, 'a')
	for line in reader:
		if line.startswith('#'):
			continue
		row = line.rstrip('\n').split('\t')
		row = formatRow(row)
		if row: writer.write(row)
	writer.close()
	reader.close()
else:
	match = lambda r1, r2: TsvJoin.compare(
		(chrorder[r1[0]], int(r1[1])),
		(chrorder[r2[0]], int(r2[1]))
	)
	def do(out, r1, r2):
		row = formatRow(r2, r1[2])
		if row: out.write(row)
	tsvjoin.join(do = do, match = match, outfile = outfile, outopts = dict(append = True))




from os import makedirs, path
from shutil import rmtree
from pyppl import Box
from bioprocs.utils import cmdargs, runcmd, mem2

tmpdir    = path.join ({{ args.tmpdir | quote}}, "{{proc.id}}.{{i.infiles[0] | fn2}}.{{job.index}}")
if not path.exists (tmpdir): makedirs (tmpdir)

params = {{args.params}}
try:
{% case args.tool %}
	############# picard
	{% when 'picard' %}
	mem = mem2({{ args.mem | quote }}, 'java')
	infiles = {{ i.infiles }}
	for i, infile in enumerate(infiles):
		params['I' + ' ' * i] = infile
	{% if args.nthread > 1 %}
	params['USE_THREADING'] = 'true'
	{% else %}
	params['USE_THREADING'] = 'false'
	{% endif %}
	params['TMP_DIR'] = tmpdir
	params['O']       = {{o.outfile | quote}}
	params['AS']      = 'true'

	cmd = '{{args.picard}} MergeSamFiles %s -Djava.io.tmpdir="%s" %s' % (mem, tmpdir, cmdargs(params, dash = '', equal = '='))
	runcmd (cmd)

	############# bamutil
	{% when 'bamutil' %}
	infiles = {{ i.infiles }}
	for i, infile in enumerate(infiles):
		params['i' + ' ' * i] = infile
	params['o'] = {{o.outfile | quote}}

	cmd = '{{args.bamutil}} mergeBam %s' % cmdargs(params)
	runcmd (cmd)

	############# samtools
	{% when 'samtools' %}
	inlist = path.join({{job.outdir | quote}}, 'bamlist.txt')
	with open(inlist, 'w') as f:
		f.write('\n'.join({{i.infiles}}) + '\n')
	params['@'] = {{args.nthread}}
	params['O'] = 'bam'
	params['b'] = inlist

	cmd = '{{args.samtools}} merge %s {{o.outfile | quote}}' % cmdargs(params)
	runcmd (cmd)

	############# sambamba
	{% when 'sambamba' %}
	params['t'] = {{args.nthread}}
	cmd = '{{args.sambamba}} merge %s {{o.outfile | quote}} {{ i.infiles | asquote }}' % cmdargs(params)
	runcmd (cmd)

{% endcase %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	rmtree (tmpdir)



from pyppl import Box
from bioprocs.utils import mem2, shell2 as shell
from pathlib import Path

fq1         = {{i.fq1 | quote}}
fq2         = {{i.fq2 | quote}}
outfq1      = {{o.outfq1 | quote}}
outfq2      = {{o.outfq2 | quote}}
tool        = {{args.tool | quote}}
cutadapt    = {{args.cutadapt | quote}}
skewer      = {{args.skewer |quote}}
trimmomatic = {{args.trimmomatic | quote}}
params      = {{args.params | repr}}
nthread     = {{args.nthread | repr}}
gz          = {{args.gz | repr}}
mem         = {{args.mem | quote }}
minlen      = {{args.minlen |repr}}
minq        = {{args.minq |repr}}
cut5        = {{args.cut5 |repr}}
cut3        = {{args.cut3 |repr}}
adapter1    = {{args.adapter1 |quote}}
adapter2    = {{args.adapter2 |quote}}
jobout      = Path({{job.outdir |quote}})

shell.load_config(
	cutadapt    = cutadapt,
	skewer      = skewer,
	trimmomatic = trimmomatic,
)

def _seqrev (seq):
	d = {
		'A':'T', 'T':'A', 'G':'C', 'C':'G',
		'a':'t', 't':'a', 'g':'c', 'c':'g'
	}
	return ''.join([d[s] for s in seq])

def run_trimmomatic():
	global mem, minlen
	mem    = mem2(mem, 'java')
	minlen = minlen * 2
	adfile = jobout / "adapters.fa"
	adfile.write_text("\n".join([
		">PE1", _seqrev(adapter1),
		">PE1_rc", adapter1,
		">PE2", _seqrev(adapter2),
		">PE2_rc", adapter2,
	]))
	params.nthread = nthread
	params._prefix = '-'
	shell.fg.trimmomatic(
		mem, "PE", params, fq1, fq2, outfq1, "/dev/null", outfq2, "/dev/null",
		"ILLUMINACLIP:%s:2:30:10" % adfile,
		"LEADING:%s" % cut5,
		"TRAILING:%s" % cut3,
		"SLIDINGWINDOW:4:%s" % minq,
		"MINLEN:%s" % minlen
	)

def run_cutadapt():
	params.a = adapter1
	params.A = adapter2
	params.u = [cut5, -cut3]
	params.U = [cut5, -cut3]
	params.m = minlen
	params.q = "{0}.{0}".format(minq)
	params.o = outfq1
	params.O = outfq2
	params._ = [fq1, fq2]
	shell.fg.cutadapt(**params)

def run_skewer():
	params.m = 'pe'
	params.t = nthread
	params.x = adapter1
	params.y = adapter2
	params.Q = minq
	params.l = minlen
	params.z = gz
	params.o = jobout / 'tmp'
	params._ = [fq1, fq2]
	shell.fg.skewer(**params)
	out1 = jobout / 'tmp-trimmed-pair1.fastq'
	out2 = jobout / 'tmp-trimmed-pair2.fastq'
	if gz:
		out1 = str(out1) + '.gz'
		out2 = str(out2) + '.gz'
	shell.mv(out1, outfq1)
	shell.mv(out2, outfq2)

TOOLS = dict(
	trimmomatic = run_trimmomatic,
	cutadapt    = run_cutadapt,
	skewer      = run_skewer
)

try:
	TOOLS[tool]()
except KeyError:
	raise ValueError('{} not supported yet.'.format(tool))




from os import path, remove, makedirs
from sys import stderr, exit
from time import sleep
from datetime import date

from pyppl import Box
from bioprocs.utils import runcmd, cmdargs, log2pyppl, logger
from bioprocs.utils.poll import Poll

gz       = {{args.gz | lambda x: bool(x)}}
infile   = {{i.infile | quote}}
infn     = {{i.infile | fn | quote}}
outfile  = {{o.outfile | quote}}
outdir   = {{o.outdir | quote}}
workdir  = {{proc.workdir | quote}}
nthread  = {{args.nthread | repr}}
cnvnator = {{args.cnvnator | repr}}
cnvkit   = {{args.cnvkit | quote}}
wandy    = {{args.wandy | quote}}
ref      = {{args.ref | quote}}
genome   = {{args.genome | quote}}
params   = {{args.params}}

def log2log(*args, **kwargs):
	# debug = False
	# pass 
	logger.info(*args, **kwargs)
	# log2pyppl(*args, **kwargs)

cnvkitParamsKeys = [
	'access', 'target', 'coverage', 'reference', 'fix', 'segment', 'call', 'plot', 'breaks',
	'gainloss', 'metrics', 'segmetrics', 'export', 'scatter', 'heatmap', 'diagram'
]

if gz:	outfile = outfile[:-3]

poll = Poll(workdir, {{proc.size}}, {{job.index}})

############## cnvkit
{% if args.tool == 'cnvkit' %}

#log2log('CNVkit: Initialize ...')

targetCov     = "{outdir}/{infn}.targetcov.cnn".format(outdir = outdir, infn = infn)
accessfile    = "{workdir}/1/output/cnvkit_access.bed".format(workdir = workdir)
targetfile    = "{workdir}/1/output/cnvkit_targets.bed".format(workdir = workdir)
refcnn        = "{workdir}/1/output/reference.cnn".format(workdir = workdir)
fixedCnr      = "{outdir}/{infn}.cnr".format(outdir = outdir, infn = infn)
segfile       = "{outdir}/{infn}.cns".format(outdir = outdir, infn = infn)
callfile      = "{outdir}/{infn}.call.cns".format(outdir = outdir, infn = infn)
# report files
breaksfile    = "{outdir}/{infn}.breaks.txt".format(outdir = outdir, infn = infn)
gainlossfile  = "{outdir}/{infn}.gainloss.txt".format(outdir = outdir, infn = infn)
metricsfile   = "{outdir}/{infn}.metrics.txt".format(outdir = outdir, infn = infn)
segmetricsfile= "{outdir}/{infn}.segmetrics.txt".format(outdir = outdir, infn = infn)
openblas_nthr = "export OPENBLAS_NUM_THREADS={nthread}; export OMP_NUM_THREADS={nthread}; export NUMEXPR_NUM_THREADS={nthread}; export MKL_NUM_THREADS={nthread}; ".format(nthread = nthread)

cnvkitAccessParams      = params.access
cnvkitAccessParams['o'] = accessfile
cmd1 = openblas_nthr + '{cnvkit} access {ref} {args}'.format(cnvkit = cnvkit, ref = repr(ref), args = cmdargs(cnvkitAccessParams))

cnvkitTargetParams = params.target
cnvkitTargetParams['o'] = targetfile
cmd2 = '{cnvkit} target {accessfile} {args}'.format(cnvkit = cnvkit, accessfile = repr(accessfile), args = cmdargs(cnvkitTargetParams))

log2log('CNVkit: Run access and target at job #0 ...')
poll.first(cmd1 + '; ' + cmd2, lockfile = 'access.poll.lock')
log2log('CNVkit: Run access and target at job #0 ... done')

cnvkitCoverageParams = params.coverage
cnvkitCoverageParams['p'] = nthread
cnvkitCoverageParams['o'] = targetCov
cmd = openblas_nthr + '{cnvkit} coverage {infile} {targetfile} {args}'.format(cnvkit = cnvkit, infile = repr(infile), targetfile = repr(targetfile), args = cmdargs(cnvkitCoverageParams))

log2log('CNVkit: Run coverage at all jobs ...')
poll.all(cmd, lockfile = 'coverage.poll.lock')
log2log('CNVkit: Run coverage at all jobs ... done')

cnvkitReferenceParams = params.reference
cnvkitReferenceParams['o'] = refcnn
cnvkitReferenceParams['f'] = ref
cmd = openblas_nthr + '{cnvkit} reference {workdir}/*/output/*/*.targetcov.cnn {args}'.format(cnvkit = cnvkit, workdir = repr(workdir), args = cmdargs(cnvkitReferenceParams))

log2log('CNVkit: Run reference at job #0 ...')
poll.first(cmd, lockfile = 'reference.poll.lock')
log2log('CNVkit: Run reference at job #0 ... done')

mtfile = "{outdir}/cnvkit_mt".format(outdir = outdir)
open(mtfile, 'w').close()

cnvkitFixParams      = params.fix
cnvkitFixParams['o'] = fixedCnr
cmd = openblas_nthr + '{cnvkit} fix {targetCov} {mtfile} {refcnn} {args}'.format(cnvkit = cnvkit, targetCov = repr(targetCov), mtfile = repr(mtfile), refcnn = repr(refcnn), args = cmdargs(cnvkitFixParams))

log2log('CNVkit: Run fix at all jobs ...')
runcmd (cmd)
log2log('CNVkit: Run fix at all jobs ... done')

if path.getsize(fixedCnr) < 60:
	open(segfile, 'w').write('chromosome	start	end	gene	log2	depth	probes	weight\\n')
else:
	cnvkitSegmentParams = params.segment
	cnvkitSegmentParams['o'] = segfile
	cnvkitSegmentParams['p'] = nthread

	cmd = openblas_nthr + '{cnvkit} segment {args} {fixedCnr}'.format(cnvkit = cnvkit, args = cmdargs(cnvkitSegmentParams), fixedCnr = repr(fixedCnr))
	log2log('CNVkit: Run segment at all jobs ...')
	runcmd (cmd)
	log2log('CNVkit: Run segment at all jobs ... done')

if path.getsize(segfile) < 60:
	open(callfile, 'w').write('chromosome	start	end	gene	log2	cn	depth	probes	weight\\n')
else:
	cnvkitCallParams = params.call
	cnvkitCallParams['o'] = callfile
	cmd = openblas_nthr + '{cnvkit} call {args} {segfile}'.format(cnvkit = cnvkit, args = cmdargs(cnvkitCallParams), segfile = repr(segfile))
	log2log('CNVkit: Run call at all jobs ...')
	runcmd (cmd)
	log2log('CNVkit: Run call at all jobs ... done')

# reports
{% if args.report %}
cnvkitBreaksParams = params.breaks
cnvkitBreaksParams['o'] = breaksfile
cmd = openblas_nthr + '{cnvkit} breaks {fixedCnr} {callfile} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), callfile = repr(callfile), args = cmdargs(cnvkitBreaksParams))
runcmd (cmd)

cnvkitGainlossParams = params.gainloss
cnvkitGainlossParams['s'] = callfile
cnvkitGainlossParams['o'] = gainlossfile
cmd = openblas_nthr + '{cnvkit} gainloss {fixedCnr} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), args = cmdargs(cnvkitGainlossParams))
runcmd (cmd)

cnvkitMetricsParams = params.metrics
cnvkitMetricsParams['s'] = callfile
cnvkitMetricsParams['o'] = metricsfile
cmd = openblas_nthr + '{cnvkit} metrics {fixedCnr} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), args = cmdargs(cnvkitMetricsParams))
runcmd (cmd)

cnvkitSegmetricsParams = params.segmetrics
cnvkitSegmetricsParams['s'] = callfile
cnvkitSegmetricsParams['o'] = segmetricsfile
cmd = openblas_nthr + '{cnvkit} segmetrics {fixedCnr} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), args = cmdargs(cnvkitSegmetricsParams))
runcmd (cmd)
{% endif %}

# plots
{% if args.plot %}
# require job.index == 0?
if not isinstance(params.scatter, list):
	params.scatter = [params.scatter]
for i, param in enumerate(params.scatter):
	param['s'] = callfile
	param['o'] = "{outdir}/{infn}.scatter{i}.pdf".format(outdir = outdir, infn = infn, i = i+1)
	cmd = openblas_nthr + '{cnvkit} scatter {fixedCnr} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), args = cmdargs(param))
	runcmd (cmd)

cnvkitDiagramParams = params.diagram
cnvkitDiagramParams['s'] = callfile
cnvkitDiagramParams['o'] = "{outdir}/{infn}.diagram.pdf".format(outdir = outdir, infn = infn)
cmd = openblas_nthr + '{cnvkit} diagram {fixedCnr} {args}'.format(cnvkit = cnvkit, fixedCnr = repr(fixedCnr), args = cmdargs(cnvkitDiagramParams))
runcmd (cmd)

poll.all(lambda x: not path.exists(x), callfile, lockfile = 'callfile.poll.lock')

{% if job.index == 0 %}
if not isinstance(params.heatmap, list):
	params.heatmap = [params.heatmap]
for i, param in enumerate(params.heatmap):
	param['o'] = "{outdir}/{infn}.heatmap{i}.pdf".format(outdir = outdir, infn = infn, i = i+1)
	cmd = openblas_nthr + '{cnvkit} heatmap {workdir}/*/output/*/*.call.cns {args}'.format(cnvkit = cnvkit, workdir = repr(workdir), args = cmdargs(param))
	runcmd (cmd)
{% endif %}
{% endif %}

# to vcf
if path.getsize(callfile) < 60: # no data generated
	with open (outfile, 'w') as fout:
		fout.write('##fileformat=VCFv4.2\\n')
		fout.write('##fileDate=%s\\n' % str(date.today()).replace('-', ''))
		fout.write('##source=CNVkit\\n')
		fout.write('##INFO=<ID=CIEND,Number=2,Type=Integer,Description="Confidence interval around END for imprecise variants">\\n')
		fout.write('##INFO=<ID=CIPOS,Number=2,Type=Integer,Description="Confidence interval around POS for imprecise variants">\\n')
		fout.write('##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the variant described in this record">\\n')
		fout.write('##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description="Imprecise structural variation">\\n')
		fout.write('##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Difference in length between REF and ALT alleles">\\n')
		fout.write('##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">\\n')
		fout.write('##INFO=<ID=FOLD_CHANGE,Number=1,Type=Float,Description="Fold change">\\n')
		fout.write('##INFO=<ID=FOLD_CHANGE_LOG,Number=1,Type=Float,Description="Log fold change">\\n')
		fout.write('##INFO=<ID=PROBES,Number=1,Type=Integer,Description="Number of probes in CNV">\\n')
		fout.write('##ALT=<ID=DEL,Description="Deletion">\\n')
		fout.write('##ALT=<ID=DUP,Description="Duplication">\\n')
		fout.write('##ALT=<ID=CNV,Description="Copy number variable region">\\n')
		fout.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\\n')
		fout.write('##FORMAT=<ID=GQ,Number=1,Type=Float,Description="Genotype quality">\\n')
		fout.write('##FORMAT=<ID=CN,Number=1,Type=Integer,Description="Copy number genotype for imprecise events">\\n')
		fout.write('##FORMAT=<ID=CNQ,Number=1,Type=Float,Description="Copy number genotype quality for imprecise events">\\n')
		fout.write('#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	{infn}\\n'.format(infn = infn))
else:
	cnvkitExportParams = params.export
	cnvkitExportParams['o'] = outfile
	cmd = openblas_nthr + '{cnvkit} export vcf {callfile} {args}'.format(cnvkit = cnvkit, callfile = repr(callfile), args = cmdargs(cnvkitExportParams))
	runcmd (cmd)

############## cnvnator
{% elif args.tool == 'cnvnator' %}
rootfile = '{outdir}/{infn}.root'.format(outdir = outdir, infn = infn)
callfile = '{outdir}/{infn}}.cnvnator'.format(outdir = outdir, infn = infn)
binsize  = {{args.binsize | repr}}

cmd = '{cnvnator} -root {rootfile} -genome {genome} -tree {infile}'.format(cnvnator = cnvnator, rootfile = repr(rootfile), genome = genome, infile = repr(infile))
runcmd (cmd)

cmd = '{cnvnator} -root {rootfile} -genome {genome} -his {binsize}'.format(cnvnator = cnvnator, rootfile = repr(rootfile), genome = genome, binsize = binsize)
runcmd (cmd)

cmd = '{cnvnator} -root {rootfile} -stat {binsize}'.format(cnvnator = cnvnator, rootfile = repr(rootfile), binsize = binsize)
runcmd (cmd)

cmd = '{cnvnator} -root {rootfile} -partition {binsize}'.format(cnvnator = cnvnator, rootfile = repr(rootfile), binsize = binsize)
runcmd (cmd)

cmd = '{cnvnator} -root {rootfile} -call {binsize}'.format(cnvnator = cnvnator, rootfile = repr(rootfile), binsize = binsize)
runcmd (cmd)

cmd = 'cd {outdir}; {cnvnator2vcf} {callfile} > {outfile}'.format(outdir = repr(outdir), cnvnator2vcf = cnvnator2vcf, callfile = repr(path.basename(callfile)), outfile = repr(outfile))
runcmd (cmd)

############## wandy
{% elif args.tool == 'wandy' %}
# get Wandy tool.info
from distutils.spawn import find_executable
toolinfo   = path.join (path.dirname(find_executable(wandy)), "tool.info")
retdir     = outdir
#if not path.exists(retdir):
#	makedirs(retdir)
myToolinfo = path.join (retdir, "tool.info")
if not path.exists(toolinfo):
	stderr.write('Cannot find tool.info in wandy source directory.')
	exit(1)
with open (toolinfo) as fin, open(myToolinfo, 'w') as fout:
	toolinfos  = [line.strip() for line in fin if not line.strip().startswith('WANDY_MEM') and not line.strip().startswith('WANDY_PARALLELED_MODE')]
	toolinfos.append ('WANDY_MEM={{args.mem}}')
	toolinfos.append ('WANDY_PARALLELED_MODE=0')
	fout.write ("\n".join(toolinfos) + "\n")

params      = {key: val for key, val in  params if key not in cnvkitParamsKeys}
params['i'] = infile
params['t'] = {{args.type | repr}}
cmd = 'cd {retdir}; {wandy} {args}'.format(retdir = repr(retdir), wandy = wandy, args = cmdargs(params))
runcmd (cmd)

# TODO: convert to vcf
open(outfile, 'w').close()
{% endif %}

if gz:	runcmd ('gzip "%s"' % (outfile))



from os import path
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
from bioprocs.utils.parallel import Parallel, distribute
from bioprocs.utils.meme import MemeReader, MemeWriter

mfile1  = {{ i.mfile1 | quote}}
mfile2  = {{ i.mfile2 | quote}}
outfile = {{ o.outfile | quote}}
outdir  = {{ o.outdir | quote}}
qval    = {{ args.qval | repr}}
tomtom  = {{ args.tomtom | quote}}
params  = {{ args.params | repr}}
nthread = {{ args.nthread | repr }}
mfile2  = mfile2 or mfile1

if nthread > 1:
	reader  = MemeReader(mfile1)
	count   = sum(1 for r in reader)
	joblist = distribute(count, nthread)
	cmdps   = []
	ocdirs  = []
	reader.rewind()
	for i in range(nthread):
		qfile = path.join(outdir, 'query-{}.meme'.format(i+1))
		ocdir = path.join(outdir, 'query-{}.tomtom'.format(i+1))
		ocdirs.append(ocdir)
		writer = MemeWriter(qfile)
		writer.meta = reader.meta
		writer.writeMeta()
		for _ in range(joblist[i]):
			try:
				writer.write(reader.next())
			except StopIteration:
				break
		writer.close()
		thparams = params.copy()
		thparams[""] = [qfile, mfile2]
		thparams.thresh = qval
		thparams.oc  = ocdir
		cmdps.append((tomtom, cmdargs(thparams, dash = '-', equal = ' ')))
	reader.close()
	Parallel(nthread, raiseExc = True).run('{} {}', cmdps)

	writer = TsvWriter(outfile)
	reader = TsvReader(path.join(ocdirs[0], 'tomtom.txt'), comment = '##', cnames = lambda header: header[1:].strip().split("\t"))
	writer.cnames = reader.cnames
	writer.writeHead(lambda cnames: "#" + "\t".join(cnames))
	reader.close()
	for ocdir in ocdirs:
		reader = TsvReader(path.join(ocdir, 'tomtom.txt'), comment = '##', cnames = lambda header: header[1:].strip().split("\t"))
		for r in reader:
			writer.write(r)
		reader.close()
	writer.close()
else:
	params[""] = [mfile1, mfile2]
	params.thresh = qval
	params.oc = outdir

	cmd = '{tomtom} {params}'.format(
		tomtom = tomtom,
		params = cmdargs(params, dash = '-', equal = ' ')
	)
	runcmd(cmd)



from pyppl import Box
from bioprocs.utils import shell2 as shell

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
params  = {{args.params | repr}}
inopts  = {{args.inopts | repr}}
case    = {{args.case | repr}}
mem     = {{args.mem | repr}}
tmpdir  = {{args.tmpdir | quote}}
unique  = {{args.unique | repr}}

shell.load_config(
	sort = dict(_dupkey = True, _env = {'LANG': 'C'} if case else {'LANG': 'en_US.UTF-8'}))

params.t = params.get('t', inopts.get('delimit', "\t"))
params.T = tmpdir
params.u = unique
params.S = mem

if 'cnames' in inopts:
	inopts.skip += 1

if inopts.skip:
	shell.head(_ = infile, n = inopts.skip, _out = outfile)
	#params.__stdout = outfile
	shell.pipe.tail(_ = infile, n = '+' + str(inopts.skip + 1)) | \
		shell.sort(_out_ = outfile, **params)
else:
	params._    = infile
	params._out = outfile
	shell.sort(**params)



from pyppl import Box
from bioprocs.utils import shell2 as shell

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
ref     = {{args.ref | quote}}
picard  = {{args.picard | quote}}
params  = {{args.params | repr}}

shell.load_config(picard = picard)

shell.fg.picard.CollectOxoGMetrics(I = infile, R = ref, O = outfile, **params)



from bioprocs.utils import alwaysList
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
suffix  = {{args.suffix | repr}}
if not suffix:
	suffix = ['']
suffix = alwaysList(suffix)

reader = TsvReader(infile, skip = 1, comment = 'CDE_ID:')
writer = TsvWriter(outfile)
writer.cnames = ['Sample', 'FID', 'IID', 'PID', 'MID', 'Sex', 'Pheno']
writer.writeHead()

for r in reader:
	for suf in suffix:
		rec        = TsvRecord()
		rec.Sample = r.bcr_patient_barcode + suf
		rec.FID    = rec.Sample
		rec.IID    = rec.Sample
		rec.PID    = 0
		rec.MID    = 0
		rec.Pheno  = -9
		rec.Sex    = 1 if r.gender == 'MALE' else 2 if r.gender == 'FEMALE' else 0
		writer.write(rec)
writer.close()




from os import remove
from pyppl import Box
from bioprocs.utils import shell

infiles = {{i.infile | repr}}
params  = {{args.params | repr}}

shell.TOOLS.bedtools = {{args.bedtools | quote}}

mergedtmp       = '{{job.outdir}}/mergedtmp.bed'
mergedtmpsorted = '{{job.outdir}}/mergedtmp.sorted.bed'

shell.touch(mergedtmp)

for infile in infiles:
	shell.grep(v = '^#', _ = infile, __stdout = mergedtmp)

shell.sort(k = ['1,1', '2,2n'], _ = mergedtmp, _stdout = mergedtmpsorted)

params.i = mergedtmpsorted
params._stdout = outfile
shell.Shell(subcmd = True, dash = '-', equal = ' ').bedtools.merge(**params).run()

shell.rm_rf(mergedtmp)
shell.rm_rf(mergedtmpsorted)



from os import path
from shutil import move
from sys import stderr
from bioprocs.utils import runcmd, cmdargs
from pyppl import Box

ref  = {{args.ref | quote}}

params = {{args.params}}
fq1 = {{o.fq1 | quote}}
fq2 = {{o.fq2 | quote}}
try:
{% case args.tool %}
	{% when 'wgsim' %}
	{% if args.gz %}
	fq1 = "{{o.fq1 | [:-3]}}"
	fq2 = "{{o.fq2 | [:-3]}}"
	{% endif %}

	params['N'] = {{args.num}}
	params['1'] = {{args.len1}}
	params['2'] = {{args.len2}}
	params['S'] = {{i.seed | lambda x: -1 if x is None else x}}
	cmd = '{{args.wgsim}} %s "%s" "%s" "%s"' % (cmdargs(params), ref, fq1, fq2)
	runcmd (cmd)

	{% if args.gz %}
	runcmd ('gzip "%s"' % fq1)
	runcmd ('gzip "%s"' % fq2)
	{% endif %}
	
	{% when 'dwgsim' %}
	prefix = {{o.fq1 | [:-8] | quote}}
	{% if args.gz %}
	fq1 = "{{o.fq1 | [:-3]}}"
	fq2 = "{{o.fq2 | [:-3]}}"
	prefix = "{{o.fq1 | [:-11]}}"
	{% endif %}

	params['N'] = {{args.num}}
	params['1'] = {{args.len1}}
	params['2'] = {{args.len2}}
	params['z'] = {{i.seed | lambda x: -1 if x is None else x}}
	params['S'] = 2
	cmd = '{{args.dwgsim}} %s "%s" "%s"' % (cmdargs(params), ref, prefix)
	runcmd (cmd)

	if path.exists (prefix + '.bwa.read1.fastq.gz'):
		move (prefix + '.bwa.read1.fastq.gz', prefix + '_1.fastq.gz')
		move (prefix + '.bwa.read2.fastq.gz', prefix + '_2.fastq.gz')
		{% if not args.gz %}
		runcmd ('gunzip "%s"' % (prefix + '_1.fastq.gz'))
		runcmd ('gunzip "%s"' % (prefix + '_2.fastq.gz'))
		{% endif %}
	else:
		move (prefix + '.bwa.read1.fastq', prefix + '_1.fastq')
		move (prefix + '.bwa.read2.fastq', prefix + '_2.fastq')
		{% if args.gz %}
		runcmd ('gzip "%s"' % (prefix + '_1.fastq'))
		runcmd ('gzip "%s"' % (prefix + '_2.fastq'))
		{% endif %}
{% endcase %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise


"""Microarray data analysis"""
import re
from os import path
from glob import glob
from pyppl import Proc, Box
#from .utils import plot, txt, dirnamePattern
from .rnaseq import pBatchEffect, pCoexp, pExprStats
from .utils import dirpat2name
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pCELDir2Matrix():
	"""
	@name:
		pCELDir2Matrix
	@description:
		Convert CEL files to expression matrix
		File names will be used as sample names (colnames)
	@input:
		`indir:file`:  the directory containing the CEL files, could be gzipped
			- If you have files, then use `pFiles2Dir` first
		`sifile:File`: the sample infor file, extensions for samples are not necessary.
			- So that it can be also used by `pMArrayDEG`
	@output:
		`outfile:file`: the expression matrix file
		`outdir:dir`:   the directory containing expr file and plots
	@args:
		`pattern`  : The pattern to filter files. Default `'*'`
		`norm`     : The normalization method. Default: rma (mas5)
		`transfm`  : The extra tranformer for the expression values after the nomralization.
			- `Note the expression values have been done with log`
		`cdffile`  : The cdffile. Default: ''
		`fn2sample`: The transformer to transform file name
	"""
	pCELDir2Matrix                  = Proc(desc = 'Merge expression files to a matrix.')
	pCELDir2Matrix.input            = "indir:file, sifile:file"
	pCELDir2Matrix.output           = "outfile:file:{{i.indir, args.pattern | *dirpat2name}}.expr.txt"
	pCELDir2Matrix.lang             = params.Rscript.value
	pCELDir2Matrix.args.fn2sample   = 'function(fn) unlist(strsplit(fn, ".", fixed = T))[1]'
	pCELDir2Matrix.args.pattern     = '*'
	pCELDir2Matrix.args.norm        = 'rma' # mas5
	pCELDir2Matrix.args.transfm     = None # mas5
	pCELDir2Matrix.args.cdffile     = ''
	pCELDir2Matrix.envs.rimport     = rimport
	pCELDir2Matrix.envs.dirpat2name = dirpat2name
	pCELDir2Matrix.script           = "file:scripts/marray/pCELDir2Matrix.r"
	return pCELDir2Matrix

@procfactory
def _pMArrayDEG():
	"""
	@name:
		pMArrayDEG
	@description:
		Detect DEGs from microarray data.
	@input:
		`efile:file`: The expression matrix file
		`gfile:file`: The sample information file
	@output:
		`outfile:file`: The file with DEGs
		`outdir:dir`  : The directory containing files and plots
	@args:
		`tool`    : The tool to use. Default: `limma`
		`annofile`: The annotation file for the probes. Default: ``
			- If not provided, raw probe name will be used.
		`filter`: The filter of the data. Default: `[0, 0]`
			- 1st element: the `rowSums` of the expression matrix
			- 2nd element: how many samples(columns) have to reach the `rowSums` given by the 1st element
		`pval`  : The pval cutoff for DEGs. Default: `0.05`
			- You may specify which kind of measurement to use
			- `p:0.05`: using cutoff 0.05 for pvalue
			- `q:0.05`: using cutoff 0.05 for qvalue(fdr)
			- If no prefix is used, then it defaults to pvalue
		`plot`  : What kind of plots to generate.
			- `mdsplot` : The MDS plot
			- `volplot` : The volcano plot
			- `maplot`  : The MA plot
			- `heatmap` : The heatmap. (use `args.hmrows` to determine how many genes to plot)
		`ggs`   : The extra ggplot element for each plot (should be able to concatenate by `+`).
			- `maplot`  : for MA plot
			- `heatmap` : for heatmap. Default: `Box(theme = {'axis.text.y': 'r:element_blank()'})`
			- `volplot` : for volcano plot
		`devpars`: The parameters for plotting device. Default: `Box(res = 300, width = 2000, height = 2000)`
	@requires:
		`r-limma`
	"""
	return pMArrayDEG

@procfactory
def _pMArrayDEG():
	"""
	@name:
		pMArrayDEG
	"""
	pMArrayDEG        = Proc(desc = 'Detect DEGs from microarray data.')
	pMArrayDEG.input  = "efile:file, gfile:file"
	pMArrayDEG.output = [
		"outfile:file:{{i.efile | fn2}}-{{i.gfile | fn2}}-DEGs/{{i.efile | fn2}}-{{i.gfile | fn2}}.degs.txt",
		"outdir:dir:{{i.efile | fn2}}-{{i.gfile | fn2}}-DEGs"
	]
	pMArrayDEG.args.tool     = 'limma'
	pMArrayDEG.args.annofile = ''
	pMArrayDEG.args.filter   = [0, 0]
	pMArrayDEG.args.pval     = 0.05
	pMArrayDEG.args.hmrows   = 100
	pMArrayDEG.args.plot     = Box(
		mdsplot = True,
		volplot = Box(fccut = 2, pcut = 0.05),
		maplot  = False,
		heatmap = False
	)
	pMArrayDEG.args.ggs = Box(
		maplot  = Box(),
		heatmap = Box(theme = {'axis.text.y': 'r:element_blank()'}),
		volplot = Box(ylab = {0: '-log10(p-value)'})
	)
	pMArrayDEG.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pMArrayDEG.envs.rimport = rimport
	pMArrayDEG.lang         = params.Rscript.value
	pMArrayDEG.script       = "file:scripts/marray/pMArrayDEG.r"
	return pMArrayDEG




from eutils import Client
from bioprocs.utils.tsvio import TsvWriter, TsvRecord
from time import sleep

term     = {{i.term | repr}}
outfile  = {{o.outfile | quote}}
prog     = {{args.prog | quote}}
apikey   = {{args.apikey | quote}} or None
db       = {{args.db | quote}}
joiner   = {{args.joiner | quote}}
record   = {{args.record}}
writer   = TsvWriter(outfile)
sleepsec = {{args.sleep | repr}}

def writerResults(rets):
	keys = []
	if rets:
		ret  = rets[0]
		keys = [key for key in dir(ret) if not key.startswith('_')]
		writer.meta.add(*keys)
		writer.writeHead()
	for ret in rets:
		r = TsvRecord()
		for key in keys:
			try:
				r[key] = getattr(ret, key)
			except:
				r[key] = ''
			if isinstance(r[key], (list, tuple)):
				try:
					r[key] = joiner.join([str(_) for _ in r[key]])
				except UnicodeEncodeError:
					r[key] = joiner.join([_.encode('utf-8') for _ in r[key]])
			else:
				try:
					r[key] = str(r[key])
				except UnicodeEncodeError:
					r[key] = r[key].encode('utf-8')
					
		if callable(record):
			r = record(r)
		elif record is not None:
			raise ValueError('Unknown record transform function (args.record).')
		if r:
			writer.write(r)

client = Client(api_key = apikey)
if prog == 'esearch':
	sret  = client.esearch(db = db, term = term)
	try:
		error = list(sret._xml_root.find('ErrorList').iterchildren())
	except:
		error = None
	
	print sret.count if not error else 0
	
	if not sret.ids:
		rets = []
	else:
		rets = client.efetch(db = db, id = sret.ids)
		rets = list(iter(rets))
	writerResults(rets)

else:
	fetches = client.efetch(db = db, id = term)
	key     = [key for key in dir(fetches) if not key.startswith('_')][0]
	rets    = getattr(fetches, key)
	writerResults(rets)

sleep (sleepsec)


from os import path
from pyppl import Box
from bioprocs.utils import shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
params   = {{args.params | repr}}
bedtools = {{args.bedtools | quote}}
gsize    = {{args.gsize | quote}}

if not gsize:
	raise ValueError('Genome size file is required (args.gsize).')
if not path.isfile(gsize):
	raise ValueError('Genome size file does not exist (args.gsize).')

bedtools = shell.Shell({'bedtools': bedtools}, subcmd = True, equal = ' ', dash = '-').bedtools

params.i = infile
params.g = gsize
params._stdout = outfile
bedtools.shift(**params).run()


from os import path
from pyppl import Box
from bioprocs.utils import shell, mem2
from bioprocs.utils.shell import Shell
from bioprocs.utils.reference import bamIndex

{% python from os import path %}

ref       = {{args.ref | quote}}
jobindex  = {{job.index | repr}}
tmpdir    = {{args.tmpdir | quote}}
procid    = {{proc.id | quote}}
infile    = {{i.infile | quote}}
prefix    = {{i.infile | fn2 | quote}}
tool      = {{args.tool | quote}}
gz        = {{args.gz | bool}}
outfile   = {{o.outfile | quote}}
params    = {{args.params | repr}}
samtools  = {{args.samtools | quote}}
mem       = {{args.mem | quote}}
nthread   = {{args.nthread | repr}}
gatk      = {{args.gatk | quote}}
cfgParams = {{args.cfgParams | repr}}
platypus  = {{args.platypus | quote}}
ssniffer  = {{args.snvsniffer | quote}}
strelka   = {{args.strelka | quote}}
vardict   = {{ args.vardict | quote}}
joboutdir = {{job.outdir | quote}}
bamIndex(infile,  samtools = None, nthread = nthread)
shell.TOOLS.update(dict(
	gatk     = gatk,
	platypus = platypus,
	ssniffer = ssniffer,
	strelka  = strelka,
	vardict  = vardict
))

if not path.exists(ref):
	raise ValueError("Reference file not specified")

tmpdir = path.join(tmpdir, "{procid}.{prefix}.{jobindex}".format(
	jobindex = jobindex,
	procid   = procid,
	prefix  = prefix
))
if not path.exists (tmpdir):
	shell.mkdir(tmpdir, p = True)

if gz:	
	outfile = outfile[:-3]

def run_gatk():
	gatkmem = mem2(mem, 'jdict')
	gatkmem['Djava.io.tmpdir={!r}'.format(tmpdir)] = True

	gatksh     = Shell(equal = ' ', dash = '-')
	params.T   = 'HaplotypeCaller'
	params.R   = ref
	params.I   = infile
	params.o   = outfile
	params.nct = nthread
	params._   = list(gatkmem.keys())

	gatksh(**params).run()
	if gz: shell.gzip(outfile)

def run_vardict():
	params.v = True
	params.G = ref
	params.b = infile
	params._stdout = outfile
	Shell().vardict(**params).run()
	if gz: shell.gzip(outfile)

def run_snvsniffer():
	hfile = path.join(joboutdir, prefix + '.header')
	Shell(subcmd = True).samtools.view(H = True, _stdout = hfile)

	params.g = ref
	params.o = outfile
	params._ = [hfile, infile]
	Shell(subcmd = True).ssniffer.snp(**params).run()
	if gz: shell.gzip(outfile)

def run_platypus():
	params.refFile     = ref
	params.bamFiles    = infile
	params.nCPU        = nthread
	params.output      = outfile
	params.logFileName = outfile + '.platypus.log'
	Shell(subcmd = True).platypus.callVariants(**params).run()
	if gz: shell.gzip(outfile)

def run_strelka():
	# config
	cfgParams.bam            = infile
	cfgParams.referenceFasta = ref
	cfgParams.runDir         = joboutdir
	Shell().strelka(**cfgParams).run()

	# run the pipeline
	params.m = 'local'
	params.j = nthread
	params.g = mem2(mem, 'G')[:-1]
	Shell({'runWorkflow': path.join(joboutdir, 'runWorkflow.py')}).runWorkflow(**params).run()

	# mv output file to desired outfile
	ofile = path.join(joboutdir, 'results', 'variants', 'genome.S1.vcf.gz')
	shell.mv(ofile, outfile + '.gz')
	if not gz: shell.gunzip(outfile + '.gz')


tools = {
	'gatk'      : run_gatk,
	'snvsniffer': run_snvsniffer,
	'strelka'   : run_strelka,
	'platypus'  : run_platypus,
	'vardict'   : run_vardict
}

try:
	tools[tool]()
except Exception:
	raise
finally:
	shell.rm(tmpdir, r = True, f = True)




import sys
from cruzdb import Genome

# get the snps
snps = []
readopts = {{args.inopts}}
{% case True %}
{% when args.inmeta | isinstance: list %}
{{read.base.py | norepeats}}
reader = readBase({{i.infile | quote}}, **readopts)
reader.meta.add(*{{args.inmeta}})
{% when args.inmeta | isinstance: dict %}
{{read.base.py | norepeats}}
reader = readBase({{i.infile | quote}}, **readopts)
reader.meta.add(*{{args.inmeta}}.items())
{% when args.inmeta | bool %}
{{read, args.inmeta | lambda x, y: x[y]['py'] | norepeats}}
reader = locals()['read{{args.inmeta | @capitalize}}']({{i.infile | quote}}, **readopts)
{% else %}
ext = {{i.infile | ext | [1:] | quote}}
ext = ext[0].upper() + ext[1:]
reader = locals()['read' + ext]({{i.infile | quote}}, **readopts)
{% endif %}

for r in reader:
	r.START = int(r.START)
	r.END   = int(r.END)
	if r.START == r.END:
		r.END = r.START + 1
	snp = (r.CHR, r.START, r.END)
	if snp not in snps:
		snps.append(snp)

genome = {{args.genome | quote}}
g     = Genome (db=genome)
dbsnp = g.snp{{args.dbsnpver}}

{{write.bedx.py | norepeats}}
writer = writeBedx({{o.outfile | quote}})
writer.meta.add(*{{args.xcols}})
#writer.writeMeta()
writer.writeHead()
for snp in snps:
	chr, start, end = snp
	s = dbsnp.filter_by(chrom=chr, chromStart = start, chromEnd = end).first()
	'''
	snp147(
		bin=1566,
		chrom='chr8',
		chromStart=128700232L,
		chromEnd=128700233L,
		name='rs7005394',
		score=0,
		strand=u'+',
		refNCBI='T',
		refUCSC='T',
		observed='C/T',
		molType=u'genomic',
		class=u'single',
		valid=set(['by-2hit-2allele', 'by-cluster', 'by-hapmap', 'by-frequency', 'by-1000genomes']),
		avHet=0.498652,
		avHetSE=0.025923,
		func=set(['ncRNA']),
		locType=u'exact',
		weight=1L,
		exceptions=set([]),
		submitterCount=26,
		submitters='1000GENOMES,ABI,BCM-HGSC-SUB,BCM_SSAHASNP,BGI,BL,BUSHMAN,COMPLETE_GENOMICS,DDI,ENSEMBL,EVA-GONL,EVA_DECODE,EVA_GENOME_DK,EVA_UK10K_ALSPAC,EVA_UK10K_TWINSUK,GMI,HAMMER_LAB,HGSV,HUMANGENOME_JCVI,ILLUMINA-UK,JMKIDD_LAB,PJP,SSAHASNP,SSMP,TISHKOFF,WEILL_CORNELL_DGM,',
		alleleFreqCount=2,
		alleles='C,T,',
		alleleNs='2634.000000,2374.000000,',
		alleleFreqs='0.525958,0.474042,',
		bitfields=set(['maf-5-all-pops', 'maf-5-some-pop'])
	)
	'''
	if not s:
		sys.stderr.write('pyppl.log.warning: Cannot find snp at: %s\n' % str(snp))
	else:
		r        = readRecord()
		r.CHR    = s.chrom
		r.START  = str(s.chromStart)
		r.END    = str(s.chromEnd)
		r.NAME   = s.name
		r.SCORE  = str(s.score)
		r.STRAND = s.strand
		for xcol in {{args.xcols}}:
			setattr(r, xcol, str(getattr(s, xcol)))
		writer.write(r)
		


from os import path
from pyppl.utils import alwaysList
from bioprocs.utils.tsvio2 import TsvReader, TsvRecord

class SampleInfoException(Exception):
	pass

class SampleInfo (object):

	NORMAL_GROUP = ['NORMAL', 'HEALTH', 'BEFORE', 'BLOOD', 'PRE', 'CONTROL']
	TUMOR_GROUP  = ['TUMOR', 'DISEASE', 'AFTER', 'TISSUE', 'POST', 'TREATMENT']

	def __init__(self, sfile):
		reader    = TsvReader(sfile)
		self.samcol = reader.cnames[0]
		if self.samcol == 'ROWNAMES':
			self.samcol = 'Sample'
			reader.cnames[0] = 'Sample'

		self.data = reader.dump()
		self.nrow = len(self.data)
		self.ncol = len(reader.cnames)
		self.colnames = reader.cnames
		self.rownames = [row[self.samcol] for row in self.data]

		expectColnames = ['Sample', 'Patient', 'Group', 'Batch']
		if not set(expectColnames) & set(self.colnames):
			raise SampleInfoException('Unexpected column names: %s.' % str(self.colnames))

	def select(self, sample = None, patient = None, group = None, batch = None, get = None):
		"""
		Select records by different standards
		"""
		ret = []
		for row in self.data:
			if sample and isinstance(sample, list) and row[self.samcol] not in sample:
				continue
			if sample and not isinstance(sample, list) and row[self.samcol] != sample:
				continue
			if patient and isinstance(patient, list) and row.Patient not in patient:
				continue
			if patient and not isinstance(patient, list) and row.Patient != patient:
				continue
			if group and isinstance(group, list) and row.Group not in group:
				continue
			if group and not isinstance(group, list) and row.Group != group:
				continue
			if batch and isinstance(batch, list) and row.Batch not in batch:
				continue
			if batch and not isinstance(batch, list) and row.Batch != batch:
				continue
			if not get:
				ret.append(row)
			else:
				if not isinstance(get, list):
					get = [get]
				r = TsvRecord()
				for g in get:
					if g == 'Sample':
						g = self.samcol
					r[g] = row[g]

				ret.append(r)
		return ret

	def toChannel(self, datadir, paired = False, raiseExc = True):
		if not paired:
			samples = self.select(get = 'Sample')
			return [path.join(datadir, r.Sample) for r in samples]
		else:
			data     = self.select(get = ['Sample', 'Patient'])
			patients = [r.Patient for r in data]
			patients = sorted(set(patients), key = patients.index)
			ret      = []
			for patient in patients:
				records = self.select(patient = patient, get = ['Sample', 'Group'])
				if len(records) != 2:
					if raiseExc:
						raise SampleInfoException('Expect 2 samples for paired channel for patient: %s' % patient)
					else:
						continue
				record1, record2 = records
				# make sure record1 is tumor and record2 is normal
				if  record1.Group.upper() in SampleInfo.TUMOR_GROUP + SampleInfo.NORMAL_GROUP and \
					record2.Group.upper() in SampleInfo.TUMOR_GROUP + SampleInfo.NORMAL_GROUP:
					if record1.Group.upper() in SampleInfo.NORMAL_GROUP:
						record1, record2 = record2, record1
				else:
					common = path.commonprefix([record1.Sample, record2.Sample])
					if common and record1.Sample < record2.Sample:
						record1, record2 = record2, record1
				ret.append((path.join(datadir, record1.Sample), path.join(datadir, record2.Sample)))
			return ret

	def dataframe(self, data = None):
		data = data or self.data
		from pandas import DataFrame
		rownames  = []
		dataframe = {}
		for row in data:
			for key, val in row.items():
				if key == self.samcol:
					rownames.append(val)
				elif key in dataframe:
					dataframe[key].append(val)
				else:
					dataframe[key] = [val]
		return DataFrame(data = dataframe, index = rownames)


class SampleInfo2(object):

	def toChannel(self, datadir, paired = False, raiseExc = True):
		if not paired:
			samples = self.getSamples()
			return [path.join(datadir, sample) for sample in samples]

		paired_samples =  self.getPairedSamples()
		ret = []
		for sample1, sample2 in paired_samples:
			if paired is not True and paired.upper() not in sample1.upper():
				sample1, sample2 = sample2, sample1
			ret.append((path.join(datadir, sample1), path.join(datadir, sample2)))
		return ret

	def _read(self, sifile):
		standard_cnames = ["", "Sample", "Patient", "Group", "Batch"]
		reader          = TsvReader(sifile)

		self.cnames     = reader.cnames
		if not self.cnames:
			raise SampleInfoException('Headers for sample information file is required.')

		if any(cname not in standard_cnames for cname in  self.cnames):
			raise SampleInfoException('Headers should be a subset of {!r}'.format(', '.join(standard_cnames)))

		if "" in self.cnames:
			self.cnames[self.cnames.index("")] = "Sample"

		self.mat = reader.dump()

	def __init__(self, sifile, checkPaired = False):
		self.mat    = None
		self.cnames = []
		self._read(sifile)
		if checkPaired and "Patient" in self.cnames:
			for patient in self.allPatients():
				if len(self.getSamples(by = 'Patient', value = patient)) != 2:
					raise SampleInfoException('Expect paired comparisons, but Patient {!r} has # samples other than 2.'.format(patient))

	def allPatients(self):
		if 'Patient' not in self.cnames:
			return None
		patients = [r.Patient for r in self.mat]
		ret = []
		for patient in patients:
			if patient not in ret:
				ret.append(patient)
		return ret

	def allGroups(self):
		allgroups = [r.Group for r in self.mat]
		ret = []
		for group in allgroups:
			if group not in ret:
				ret.append(group)
		return ret

	def allSamples(self, unique = False, datadir = None):
		samples = self.getSamples()
		if unique:
			ret = []
			for sample in samples:
				if sample not in ret:
					ret.append(sample)
			return ret
		return [path.join(datadir, sample) for sample in samples] if datadir else samples

	def getSamples(self, by = None, value = None, returnAll = False):
		if by and by not in self.cnames:
			raise SampleInfoException('{!r} is not a valid column name.'.format(by))
		if not by:
			return [r if returnAll else r.Sample for r in self.mat]
		return [r if returnAll else r.Sample for r in self.mat if r[by] == value]

	def sampleInfo(self, sample, info = None):
		if not info:
			return [r for r in self.mat if r.Sample == sample]
		if info not in self.cnames:
			raise SampleInfoException('{!r} is not a valid column name.'.format(info))
		return [r[info] for r in self.mat if r.Sample == sample]

	def select(self, samples):
		self.mat = [r for r in self.mat if r.Sample in samples]

	def isPaired(self):
		allpatients = self.allPatients()
		if not allpatients:
			return False
		for patient in allpatients:
			if len(self.getSamples(by = 'Patient', value = patient)) != 2:
				return False
		return True

	def getPairedSample(self, sample):
		patient = self.sampleInfo(sample, info = 'Patient')[0]
		samples = self.getSamples(by = 'Patient', value = patient)
		return samples[1 - samples.index(sample)]

	def getPairedSamples(self, datadir = None):
		patients = self.allPatients()
		groups   = self.allGroups()

		ret = []
		for patient in patients:
			samples = self.getSamples(by = 'Patient', value = patient, returnAll = True)
			s1 = path.join(datadir, samples[0].Sample) if datadir else samples[0].Sample
			s2 = path.join(datadir, samples[1].Sample) if datadir else samples[1].Sample
			if samples[0].Group != groups[1]: # make sure it's (tumor, normal) pair
				ret.append((s1, s2))
			else:
				ret.append((s2, s1))
		return ret





from os import path
from sys import stderr
from cruzdb import Genome
from tempfile import gettempdir
from pyppl import Box
from bioprocs.utils import alwaysList
from bioprocs.utils.cache import Cache
from bioprocs.utils.tsvio import TsvReader, TsvWriter, TsvRecord

class RecordNotFound(Exception):
	pass

def snpinfo(infile, outfile = None, notfound = 'ignore', genome = 'hg19', dbsnpver = "150", inopts = None, outopts = None, snpcol = None, cachedir = gettempdir()):
	_inopts = Box(skip = 0, comment = '#', delimit = '\t')
	_inopts.update(inopts or {})
	inopts  = _inopts

	_outopts = Box(delimit = '\t', headDelimit = '\t', headPrefix = '', headTransform = None, head = True, ftype = 'bed', cnames = 'refUCSC, alleles, alleleFreqs, alleleFreqCount')
	_outopts.update(outopts)
	outopts  = _outopts
	cnames   = alwaysList(outopts['cnames'])

	reader = TsvReader(infile, **inopts)
	if not reader.meta: reader.autoMeta()
	snpcol = snpcol or reader.meta.keys()[0]

	snps = list(set([r[snpcol] for r in reader]))
	reader.rewind()

	dbfile = path.join(cachedir, 'snpinfo_%s_%s.db' % (genome, dbsnpver))
	schema = {
		'chrom'          : 'text',              # chr8
		'chromStart'     : 'int',               # 128700232L
		'chromEnd'       : 'end',               # 128700233L
		'name'           : 'text primary key',  # rs7005394
		'score'          : 'real',              # 0
		'strand'         : 'text',              # +
		'refNCBI'        : 'text',              # T
		'refUCSC'        : 'text',              # T
		'observed'       : 'text',              # C/T
		'class'          : 'single',            # single
		'avHet'          : 'real',              # 0.49
		'avHetSE'        : 'real',              # 0.02
		'func'           : 'text',              # set(['ncRNA'])
		'submitterCount' : 'int',               # 20
		'submitters'     : 'text',              # 1000GENOMES,ABI,BCM-HGSC-SUB,BCM_SSAHASNP,BGI,BL ...
		'alleleFreqCount': 'int',               # 2
		'alleles'        : 'text',              # C,T,
		'alleleNs'       : 'text',              # 2634.000000,2374.000000,
		'alleleFreqs'    : 'text',              # 0.525958,0.474042,
	}
	cache   = Cache(dbfile, 'snpinfo', schema, 'name')
	dummies = {
		'func': dict(
			query  = Cache.DUMMY['array']['query'],
			find   = Cache.DUMMY['array']['find'],
			insert = lambda col, data: (col, ' // ' + ' // '.join(Cache._uniqueData(list(data), True))),
			update = lambda col, data: (col, Function.concat(Field(col), value = ' // ' + ' // '.join(Cache._uniqueData(list(d), True)))),
			result = lambda data: data if isinstance(data, list) else list(filter(None, data.split(' // ')))
		)
	}
	columns = [
		'chrom', 'name', 'chromStart', 'chromEnd', 'score', 'strand'
	] + cnames
	ret, allrest = cache.query(columns, {'name': snps}, dummies)

	dbsnp = Genome(db = genome)
	dbsnp = getattr(dbsnp, "snp%s" % dbsnpver)

	writer = None
	if outfile:
		head          = outopts['head']
		headPrefix    = outopts['headPrefix']
		headDelimit   = outopts['headDelimit']
		headTransform = outopts['headTransform']
		del outopts['head']
		del outopts['headPrefix']
		del outopts['headDelimit']
		del outopts['headTransform']
		writer = TsvWriter(outfile, **outopts)
		if head:
			writer.writeHead(prefix = headPrefix, delimit = headDelimit, transform = headTransform)

	if writer:
		for r in ret.values():
			r.CHR    = r.chrom
			r.START  = r.chromStart
			r.END    = r.chromEnd
			r.NAME   = r.name
			r.SCORE  = r.score
			r.STRAND = r.strand
			writer.write(r)

	cached = []
	if allrest:
		for snp in allrest['name']:
			s = dbsnp.filter_by(name=snp).first()
			if not s:
				if notfound == 'error':
					raise RecordNotFound('Record not found: %s' % snp)
				elif notfound == 'skip':
					continue
				else:
					stderr.write('Record not found: %s \n' % snp)
					continue
			cached.append(s)
			if writer:
				r = TsvRecord()
				r.CHR    = s.chrom
				r.START  = s.chromStart
				r.END    = s.chromEnd
				r.NAME   = s.name
				r.SCORE  = s.score
				r.STRAND = s.strand
				for cname in cnames:
					setattr(r, cname, getattr(s, cname))
				writer.write(r)

	# save cached data
	cachedata = {}
	for c in cached:
		for k in schema.keys():
			if not k in cachedata:
				cachedata[k] = []
			cachedata[k].append(getattr(c, k))
	if cachedata:
		cache.save(cachedata, dummies)
	return {r.name: r for r in ret.values() + cached}



from bioprocs.utils import FileConn

{% python from bioprocs.utils import alwaysList %}
infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
dbsnp    = {{args.dbsnp | quote}}
notfound = {{args.notfound | quote}}
exist    = {{args.exist | quote}}
chrorder = {{args.chrorder | alwaysList | repr}}
chrorder = dict(zip(chrorder, list(range(len(chrorder)))))

def compare(chr1, pos1, chr2, pos2):
	"""
	return 0 if they equal, -1 if snp1 less, else 1
	"""
	chr1 = chrorder.get(chr1, chr1)
	chr2 = chrorder.get(chr2, chr2)
	pos1 = int(pos1)
	pos2 = int(pos2)
	return -1 if (chr1, pos1) < (chr2, pos2) else 1 if (chr1, pos1) > (chr2, pos2) else 0

with FileConn(infile) as fin, FileConn(dbsnp) as fdb, FileConn(outfile, 'w') as fout:
	fout.write(fin.next())
	rin = rdb = None
	while True:
		try:
			rin = rin or fin.next()
			rdb = rdb or fdb.next()
			if rdb.startswith('#'):
				rdb = None
				continue
			inparts = rin.split('\t')
			names   = inparts[0].split('_')
			if len(names) == 4:
				(chr_in, pos_in, ref, alt) = names
			else:
				(chr_in, pos_in, rs, ref, alt) = names
			if exist == 'keep' and rs.startswith('rs'):
				fout.write(rin)
				rin = None
				continue
			(chr_db, pos_db, rs_db) = rdb.split('\t')[:3]
			c = compare(chr_in, pos_in, chr_db, pos_db)
			if c < 0:
				inparts[0] = '_'.join((chr_in, pos_in, notfound, ref, alt))
				fout.write('\t'.join(inparts))
				rin = None
				continue
			if c > 0:
				rdb = None
				continue
			inparts[0] = '_'.join((chr_in, pos_in, rs_db, ref, alt))
			fout.write('\t'.join(inparts))
			rin = None
		except StopIteration:
			break
		except:
			from traceback import print_exc
			print_exc()
			rin = rdb = None
			continue



from pyppl import Box
from os import path, symlink
from bioprocs.utils.gene import genenorm

genenorm(
	infile   = {{i.infile | quote}},
	outfile  = {{o.outfile | quote}},
	notfound = {{args.notfound | quote}},
	frm      = {{args.frm | quote}},
	to       = {{args.to  | quote}},
	genome   = {{args.genome | quote}},
	inopts   = {{args.inopts}},
	outopts  = {{args.outopts}},
	genecol  = {{args.genecol | quote}},
	cachedir = {{args.cachedir | quote}}
)



from pyppl import Box
from bioprocs.utils.snp import snpinfo

infile   = {{i.snpfile | quote}}
outfile  = {{o.outfile | quote}}
genome   = {{args.genome | quote}}
dbsnpver = {{args.dbsnpver | quote}}
notfound = {{args.notfound | quote}}
inopts   = {{args.inopts}}
outopts  = {{args.outopts}}
snpcol   = {{args.snpcol | quote}}
cachedir = {{args.cachedir | quote}}

snpinfo (
	infile    = infile,
	outfile   = outfile,
	notfound  = notfound,
	genome    = genome,
	dbsnpver  = dbsnpver,
	inopts    = inopts,
	outopts   = outopts,
	snpcol    = snpcol,
	cachedir  = cachedir
)



#!/usr/bin/env python
"""Call mutatoins from 2nd-gen sequencing data"""

import sys
from pyppl import Box
from bioprocs import params

import sys
from os import path
from pyppl import PyPPL, Channel, Box
from bioprocs import params

params._prefix          = '-'
params.intype           = 'bam' # ebam, fastq
params.intype.desc      = 'The input file types. Either bam, ebam or fastq.\nEbam means the bam files need to reformatted into fastq files.'
params.muts             = ['germ'] # or ['germ', 'soma', 'scnv', 'gcnv']
params.muts.type        = list
params.muts.desc        = 'What kind of mutations to call.\nNote: soma need paired information'
params.indir.required   = True
params.indir.desc       = 'The input directory containing input files.'
params.indir.callback   = lambda opt: opt.setValue(path.realpath(opt.value))
params.aligner          = 'bwa'
params.aligner.desc     = 'The alignment tool.'
params.saminfo.required = True
params.saminfo.desc     = """The sample information file:
Column 1: the basename of the sample file in '-indir'
Column 2: Group if '-muts' includes 'soma', basically, 'TUMOR' and 'NORMAL'
Column 3: Patient if multiple samples belong to the same patient.
Example:
Sample	Group	Patient
A1.bam	TUMOR	A
A2.bam	NORMAL	A
B1.bam	TUMOR	A
B2.bam	NORMAL	A
"""
params.nthread        = 1
params.nthread.desc   = 'Use multithreading if possible.'
params.exdir.required = True
params.exdir.desc     = 'Where to export the result files.'
params.runner         = 'local'
params.runner.desc    = 'The runner to run the processes.'
params.forks          = 1
params.forks.desc     = 'How many jobs to run simultanuously.'
params.logfile        = ''
params.logfile.desc   = 'Where to save the logs.'
params.compress       = True
params.compress.desc  = 'Use gzip and bam file to save space.'
params.ppldir         = './workdir'
params.ppldir.desc    = 'The pipeline directory.'

def main():
	opts = params._parse(dict_wrapper = Box)

	from bioprocs.common import pFiles2Dir
	from bioprocs.sambam import pBam2Gmut, pBamPair2Smut
	from bioprocs.utils.sampleinfo import SampleInfo2 as SampleInfo
	from bioprocs.sets.wxs import aPrepareBam, aBam2SCNV, aBam2GCNV

	starts = []
	saminfo = SampleInfo(opts.saminfo)
	aPrepareBam.pFastq2Sam.args.tool    = opts.aligner
	if aPrepareBam.pSam2Bam.args.tool == 'elprep' and aPrepareBam.pSam2Bam.args.steps.recal:
		aPrepareBam.modules.norecal()
	if opts.compress:
		aPrepareBam.args.gz = True
		aPrepareBam.pFastq2Sam.args.outfmt = 'bam'

	pBamDir         = pFiles2Dir
	pBamDir.runner  = 'local'
	if opts.intype == 'ebam':
		#aPrepareBam.input = [Channel.fromPattern(path.join(opts.indir, '*.bam'))]
		aPrepareBam.modules.ebam(restore = False)
		aPrepareBam.input = Channel.create(saminfo.toChannel(opts.indir)).unique()

		pBamDir.depends = aPrepareBam
		pBamDir.input   = lambda ch: [ch.flatten()]

		starts.append(aPrepareBam)

	elif opts.intype in ('fq', 'fastq'):
		# pair-end fastq files
		# *.fq, *.fq.gz *.fastq, *.fastq.gz
		# sample info should be:
		# +--------------+----------+---------+
		# | Sample	     | Patient  | Group   |
		# | x_Tumor.bam	 | x        | TUMOR	  |
		# | x_Normal.bam | x        | NORMAL  |
		# | ...          | ...      | ...     |
		# +--------------+----------+---------+
		# corresponding fastq files would be:
		# x_Tumor_1.fq  / x_Tumor_1.fastq  / x_Tumor_1.fq.gz  / x_Tumor_1.fastq.gz
		# x_Tumor_2.fq  / x_Tumor_2.fastq  / x_Tumor_2.fq.gz  / x_Tumor_2.fastq.gz
		# x_Normal_1.fq / x_Normal_1.fastq / x_Normal_1.fq.gz / x_Normal_1.fastq.gz
		# x_Normal_2.fq / x_Normal_2.fastq / x_Normal_2.fq.gz / x_Normal_2.fastq.gz
		def bam2fqpair(fq):
			fqdir    = path.dirname(fq)
			bname    = path.splitext(path.basename(fq))[0]
			exts1    = ['_1.fq', '_1.fq.gz', '_1.fq.gz', '_1.fastq.gz']
			exts2    = ['_2.fq', '_2.fq.gz', '_2.fq.gz', '_2.fastq.gz']
			fqfiles1 = [path.join(fqdir, bname + ext) for ext in exts1]
			fqfile1  = [fqfile for fqfile in fqfiles1 if path.isfile(fqfile)][0]
			fqfiles2 = [path.join(fqdir, bname + ext) for ext in exts2]
			fqfile2  = [fqfile for fqfile in fqfiles2 if path.isfile(fqfile)][0]
			return fqfile1, fqfile2

		aPrepareBam.modules.fastq()
		aPrepareBam.input = [bam2fqpair(fq) for fq in saminfo.toChannel(opts.indir)]

		pBamDir.depends = aPrepareBam.pBamRecal
		pBamDir.input   = lambda ch: [ch.flatten()]

		starts.append(aPrepareBam)
	else:
		pBamDir.input = [saminfo.toChannel(opts.indir)]
		starts.append(pBamDir)

	if 'germ' in opts.muts:
		pBam2Gmut.depends      = pBamDir
		pBam2Gmut.input        = lambda ch: ch.expand(0, "*.bam")
		pBam2Gmut.exdir        = path.join(opts.exdir, 'germline')
	if 'soma' in opts.muts:
		pBamPair2Smut.depends      = pBamDir
		pBamPair2Smut.input        = lambda ch: saminfo.toChannel(ch.get(), paired = True)
		pBamPair2Smut.exdir        = path.join(opts.exdir, 'somatic')
	if 'scnv' in opts.muts:
		aBam2SCNV.modules.plots()
		aBam2SCNV.pBamDir.depends   = pBamDir
		aBam2SCNV.pSampleInfo.input = [opts.saminfo]
		aBam2SCNV.exdir             = path.join(opts.exdir, 'scnv')
		starts.append(aBam2SCNV)
	if 'gcnv' in opts.muts:
		aBam2GCNV.modules.plots()
		aBam2GCNV.pBamDir.depends   = pBamDir
		aBam2GCNV.pSampleInfo.input = [opts.saminfo]
		aBam2GCNV.exdir             = path.join(opts.exdir, 'gcnv')
		starts.append(aBam2GCNV)

	config = {
		'default': {
			'forks' : int(opts.forks),
			'ppldir': opts.ppldir,
			'args'  : {'nthread': opts.nthread},
			'_log'  : {'file': opts.logfile}
		},
	}

	PyPPL(config).start(starts).run(opts.runner)

if __name__ == "__main__":
	main()


"""
A set of processes to generate/process fastq/fasta files
"""

from os import path
from pyppl import Proc, Box
from . import params, bashimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

def _getFastqFn(fastq):
	ret = path.basename(fastq)
	if ret.endswith('.gz')   : ret = ret[:-3]
	if ret.endswith('.fastq'): ret = ret[:-6]
	if ret.endswith('.fq')   : ret = ret[:-3]
	if ret.endswith('.clean'): ret = ret[:-6]
	if ret.endswith('_clean'): ret = ret[:-6]
	return ret

def _getCommonName(f1, f2):
	ret = path.basename(path.commonprefix([f1, f2]))
	while not ret[-1].isalnum(): ret = ret[:-1]
	return ret

@procfactory
def _pFastq2Expr():
	"""
	@name:
		pFastq2Expr
	@description:
		Use Kallisto to get gene expression from pair-end fastq files.
	@input:
		`fqfile1:file`: The fastq file1.
		`fqfile2:file`: The fastq file2.
	@output:
		`outfile:file`: The expression file
		`outdir:dir`  : Output direcotry with expression and other output files
	@args:
		`params`  : Other parameters for `kallisto quant`. Default: `Box()`
		`idxfile` : The kallisto index file. Default: `params.kallistoIdx`
		`kallisto`: The path to `kallisto`. Default: `params.kallisto`
		`nthread` : # threads to use. Default: `1`
	"""
	pFastq2Expr        = Proc(desc = 'Use Kallisto to get gene expression from pair-end fastq files.')
	pFastq2Expr.input  = "fqfile1:file, fqfile2:file"
	pFastq2Expr.output = [
		"outfile:file:{{i.fqfile1, i.fqfile2 | *commonname}}/{{i.fqfile1, i.fqfile2 | *commonname}}.expr.txt",
		"outdir:dir:{{i.fqfile1, i.fqfile2 | *commonname}}"
	]
	pFastq2Expr.args.params     = Box()
	pFastq2Expr.args.idxfile    = params.kallistoIdx.value
	pFastq2Expr.args.kallisto   = params.kallisto.value
	pFastq2Expr.args.nthread    = 1
	pFastq2Expr.envs.commonname = lambda f1, f2, path = __import__('os').path: path.basename(path.commonprefix([f1, f2])).rstrip('_. ,[]')
	pFastq2Expr.envs.bashimport = bashimport
	pFastq2Expr.preCmd          = """
	{{bashimport}} reference.bash
	reference kallisto {{args.idxfile | squote}}
	"""
	pFastq2Expr.lang   = params.python.value
	pFastq2Expr.script = "file:scripts/fastx/pFastq2Expr.py"
	return pFastq2Expr

@procfactory
def _pFastqSim():
	"""
	@name:
		pFastqSim
	@description:
		Simulate reads
	@input:
		`seed`: The seed to generate simulation file
			- None: use current timestamp.
	@output:
		`fq1:file`: The first pair read file
		`fq2:file`: The second pair read file
	@args:
		`tool`:  The tool used for simulation. Default: wgsim (dwgsim)
		`len1`:  The length of first pair read. Default: 100
		`len2`:  The length of second pair read. Default: 100
		`num`:   The number of read PAIRs. Default: 1000000
		`gz`:    Whether generate gzipped read file. Default: True
		`wgsim`: The path of wgsim. Default: wgsim
		`dwgsim`:The path of wgsim. Default: dwgsim
		`ref`:   The reference genome. Required
		`params`:Other params for `tool`. Default: ""
	@requires:
		[`wgsim`](https://github.com/lh3/wgsim)
	"""
	pFastqSim             = Proc(desc = 'Simulate pair-end reads.')
	pFastqSim.input       = "seed"
	pFastqSim.output      = "fq1:file:read{{i.seed}}_1.fastq{% if args.gz %}.gz{% endif %}, fq2:file:read{{i.seed}}_2.fastq{% if args.gz %}.gz{% endif %}"
	pFastqSim.args.tool   = 'wgsim'
	pFastqSim.args.wgsim  = params.wgsim.value
	pFastqSim.args.dwgsim = params.dwgsim.value
	pFastqSim.args.len1   = 100
	pFastqSim.args.len2   = 100
	pFastqSim.args.num    = 1000000
	pFastqSim.args.gz     = False
	pFastqSim.args.params = Box()
	pFastqSim.args.ref    = params.ref.value
	pFastqSim.lang        = params.python.value
	pFastqSim.script      = "file:scripts/fastx/pFastqSim.py"
	return pFastqSim

@procfactory
def _pFastQC():
	"""
	@name:
		pFastQC
	@description:
		QC report for fastq file
	@input:
		`fq:file`:    The fastq file (also fine with gzipped)
	@output:
		`outdir:dir`: The output direcotry
	@args:
		`tool`:    The tool used for simulation. Default: fastqc
		`fastqc`:  The path of fastqc. Default: fastqc
		`nthread`: Number of threads to use. Default: 1
		`params`:Other params for `tool`. Default: ""
	@requires:
		[`fastqc`](https://www.bioinformatics.babraham.ac.uk/projects/fastqc/)
	"""
	pFastQC                 = Proc(desc = 'QC report for fastq file.')
	pFastQC.input           = "fq:file"
	pFastQC.output          = "outdir:dir:{{i.fq | getFastqFn}}.fastqc"
	pFastQC.args.tool       = 'fastqc'
	pFastQC.args.fastqc     = params.fastqc.value
	pFastQC.args.nthread    = 1
	pFastQC.args.params     = Box()
	pFastQC.lang            = params.python.value
	pFastQC.envs.getFastqFn = _getFastqFn
	pFastQC.script          = "file:scripts/fastx/pFastQC.py"
	return pFastQC

@procfactory
def _pFastMC():
	"""
	@name:
		pFastMC
	@description:
		Multi-QC based on pFastQC
	@input:
		`qcdir:file`:  The direcotry containing QC files
	@output:
		`outdir:dir`: The output direcotry
	@args:
		`tool`:    The tool used for simulation. Default: multiqc
		`multiqc`: The path of fastqc. Default: multiqc
		`params`:  Other params for `tool`. Default: ""
	@requires:
		[`multiqc`](http://multiqc.info/)
	"""
	pFastMC              = Proc(desc = 'Multi-QC based on pFastQC.')
	pFastMC.input        = "qcdir:file"
	pFastMC.output       = "outdir:dir:{{i.qcdir | fn}}.multiqc"
	pFastMC.args.tool    = 'multiqc'
	pFastMC.args.multiqc = params.multiqc.value
	pFastMC.args.params  = Box()
	pFastMC.lang         = params.python.value
	pFastMC.script       = "file:scripts/fastx/pFastMC.py"
	return pFastMC

@procfactory
def _pFastqTrim():
	"""
	@input:
		fq1: The input fastq file 1
		fq2: The input fastq file 2
	@output:
		outfq1: The trimmed fastq file 1
		outfq2: The trimmed fastq file 2
	@args:
		tool       : The tools used for trimming. Available: trimmomatic, cutadapt or skewer
		cutadapt   : The path of seqtk.
		skewer     : The path of fastx toolkit trimmer.
		trimmomatic: The path of trimmomatic.
		params     : Other params for `tool`.
		nthread    : Number of threads to be used.
			- Not for cutadapt
		gz : Whether gzip output files.
		mem: The memory to be used.
			- Only for trimmomatic
		minlen: Discard trimmed reads that are shorter than `minlen`.
			- For trimmomatic, the number will be `minlen`*2 for MINLEN, as it filters before trimming
		minq: Minimal mean qulity for 4-base window or leading/tailing reads.
		cut5: Remove the 5'end reads if they are below qulity.
		cut3: Remove the 3'end reads if they are below qulity.
			- Not for skewer
		adapter1: The adapter for sequence.
		adapter2: The adapter for pair-end sequence.
	@requires:
		[`cutadapt`](http://cutadapt.readthedocs.io/en/stable/guide.html)
		[`skewer`](https://github.com/relipmoc/skewer)
		[`trimmomatic`](https://github.com/timflutre/trimmomatic)
	"""
	return Box(
		desc   = 'Trim pair-end reads in fastq file.',
		lang   = params.python.value,
		envs   = Box(getFastqFn = _getFastqFn),
		input  = "fq1:file, fq2:file",
		output = [
			"outfq1:file:{{i.fq1 | getFastqFn}}.fastq{% if args.gz %}.gz{% endif %}",
			"outfq2:file:{{i.fq2 | getFastqFn}}.fastq{% if args.gz %}.gz{% endif %}"
		],
		args = Box(
			tool        = 'skewer',
			cutadapt    = params.cutadapt.value,
			skewer      = params.skewer.value,
			trimmomatic = params.trimmomatic.value,
			params      = Box(),
			nthread     = 1,
			gz          = False,
			mem         = params.mem4G.value,
			minlen      = 18,
			minq        = 3,
			cut5        = 3,
			cut3        = 3,
			adapter1    = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC',
			adapter2    = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTA'))

@procfactory
def _pFastqSETrim():
	"""
	@name:
		pFastqSETrim
	@description:
		Trim single-end FASTQ reads
	@input:
		`fq:file`:  The input fastq file
	@output:
		`outfq:file`: The trimmed fastq file
	@args:
		`tool`        : The tools used for trimming. Default: trimmomatic (cutadapt|skewer)
		`cutadapt`    : The path of seqtk. Default: cutadapt
		`skewer`      : The path of fastx toolkit trimmer. Default: skewer
		`trimmomatic` : The path of trimmomatic. Default: trimmomatic
		`params`      : Other params for `tool`. Default: ""
		`nthread`     : Number of threads to be used. Default: 1
		- Not for cutadapt
		`gz`          : Whether gzip output files. Default: True
		`mem`         : The memory to be used. Default: 4G
		- Only for trimmomatic
		`minlen`      : Discard trimmed reads that are shorter than `minlen`. Default: 18
		- For trimmomatic, the number will be `minlen`*2 for MINLEN, as it filters before trimming
		`minq`        : Minimal mean qulity for 4-base window or leading/tailing reads. Default: 3
		`cut5`        : Remove the 5'end reads if they are below qulity. Default: 3
		`cut3`        : Remove the 3'end reads if they are below qulity. Default: 3
		- Not for skewer
		`adapter`     : The adapter for sequence. Default: AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC
	@requires:
		[`cutadapt`](http://cutadapt.readthedocs.io/en/stable/guide.html)
		[`skewer`](https://github.com/relipmoc/skewer)
		[`trimmomatic`](https://github.com/timflutre/trimmomatic)
	"""
	pFastqSETrim                  = Proc(desc = 'Trim single-end reads in fastq file.')
	pFastqSETrim.input            = "fq:file"
	pFastqSETrim.output           = "outfq:file:{{i.fq | getFastqFn }}.fastq{% if args.gz %}.gz{% endif %}"
	pFastqSETrim.lang             = params.python.value
	pFastqSETrim.args.tool        = 'skewer'
	pFastqSETrim.args.cutadapt    = params.cutadapt.value
	pFastqSETrim.args.skewer      = params.skewer.value
	pFastqSETrim.args.trimmomatic = params.trimmomatic.value
	pFastqSETrim.args.params      = Box()
	pFastqSETrim.args.nthread     = 1
	pFastqSETrim.args.gz          = False
	pFastqSETrim.args.mem         = params.mem4G.value
	pFastqSETrim.args.minlen      = 18
	pFastqSETrim.args.minq        = 3
	pFastqSETrim.args.cut5        = 3
	pFastqSETrim.args.cut3        = 3
	pFastqSETrim.args.adapter     = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC'
	pFastqSETrim.envs.getFastqFn  = _getFastqFn
	pFastqSETrim.script           = "file:scripts/fastx/pFastqSETrim.py"
	return pFastqSETrim


@procfactory
def _pFastqSE2Sam():
	"""
	@name:
		pFastqSE2Sam
	@description:
		Cleaned paired fastq (.fq, .fq.gz, .fastq, .fastq.gz file to mapped sam/bam file
	@args:
		`tool`:   The tool used for alignment. Default: bwa (bowtie2|ngm)
		`bwa`:    Path of bwa, default: bwa
		`ngm`:    Path of ngm, default: ngm
		`bowtie2`:Path of bowtie2, default: bowtie2
		`rg`:     The read group. Default: {'id': '', 'pl': 'Illumina', 'pu': 'unit1', 'lb': 'lib1', 'sm': ''}
		- `id` will be parsed from filename with "_LX_" in it if not given
		- `sm` will be parsed from filename
		`ref`:    Path of reference file
		`params`: Other params for tool, default: ''
	"""
	pFastqSE2Sam                    = Proc(desc = 'Map cleaned single-end fastq file to reference genome.')
	pFastqSE2Sam.input              = "fq:file"
	pFastqSE2Sam.output             = "outfile:file:{{i.fq | getFastqFn }}.{{args.outfmt}}"
	pFastqSE2Sam.args.outfmt        = "sam"
	pFastqSE2Sam.args.tool          = 'bwa'
	pFastqSE2Sam.args.bwa           = params.bwa.value
	pFastqSE2Sam.args.ngm           = params.ngm.value
	pFastqSE2Sam.args.star          = params.star.value
	pFastqSE2Sam.args.samtools      = params.samtools.value
	pFastqSE2Sam.args.bowtie2       = params.bowtie2.value
	pFastqSE2Sam.args.bowtie2_build = params.bowtie2.value + '-build'
	pFastqSE2Sam.args.rg            = Box(id = '', pl = 'Illumina', pu = 'unit1', lb = 'lib1', sm = '')
	pFastqSE2Sam.args.ref           = params.ref.value
	pFastqSE2Sam.args.refgene       = params.refgene.value
	pFastqSE2Sam.args.nthread       = 1
	pFastqSE2Sam.args.params        = Box()
	pFastqSE2Sam.envs.getFastqFn    = _getFastqFn
	pFastqSE2Sam.lang               = params.python.value
	pFastqSE2Sam.script             = "file:scripts/fastx/pFastqSE2Sam.py"
	return pFastqSE2Sam

@procfactory
def _pFastq2Sam():
	"""
	@description:
		Cleaned paired fastq (.fq, .fq.gz, .fastq, .fastq.gz file to mapped sam/bam file
	@args:
		`tool`   : The tool used for alignment. Available: bowtie2, ngm or star.
		`bwa`    : Path of bwa
		`ngm`    : Path of ngm
		`star`   : Path of STAR
			- The genomeDir should be at `/path/to/hg19.star` if `args.ref` is `/path/to/hg19.fa`
		`bowtie2`: Path of bowtie2
		`rg`:     The read group.
			- `id` will be parsed from filename with "_LX_" in it if not given
			- `sm` will be parsed from filename
		`ref`    : Path of reference file
		`refexon`: The GTF file for STAR to build index. It's not neccessary if index is already been built.
		`params` : Other params for tool
	"""
	return Box(
		desc   = 'Map cleaned paired fastq file to reference genome.',
		lang   = params.python.value,
		input  = "fq1:file, fq2:file",
		output = "outfile:file:{{i.fq1, i.fq2 | path.commonprefix | bn | .rstrip: '_. ,[]' }}.{{args.outfmt}}",
		envs   = Box(path = path),
		preCmd = """
			{{bashimport}} reference.bash
			export bwa={{args.bwa | squote}}
			export ngm={{args.ngm | squote}}
			export star={{args.star | squote}}
			export samtools={{args.samtools | squote}}
			export bowtie2={{args.bowtie2 | squote}}
			export nthread={{args.nthread}}
			export refexon={{args.refexon | squote}}
			reference {{args.tool | squote}} {{args.ref | squote}}""",
		args = Box(
			tool     = 'bwa',
			outfmt   = 'sam',
			bwa      = params.bwa.value,
			ngm      = params.ngm.value,
			star     = params.star.value,
			samtools = params.samtools.value,
			bowtie2  = params.bowtie2.value,
			rg       = Box(id = '', pl = 'Illumina', pu = 'unit1', lb = 'lib1', sm = ''),
			ref      = params.ref.value,
			refexon  = params.refexon.value,
			nthread  = 1,
			params   = Box()
		)
	)




from pyppl import Box
from bioprocs.utils import shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote}}
extend   = {{ args.extend | bool}}
gsize    = {{ args.gsize | quote}}
params   = {{ args.params | repr}}
bedtools = {{ args.bedtools | quote}}

shell.TOOLS.bedtools = bedtools
bedtools = shell.Shell(subcmd = True, dash = '-', equal = ' ').bedtools

params['g']   = gsize
params['i']   = infile

if not 'l' and not 'r' and not 'b' in params:
	raise ValueError('You have to define a length to flank (args.params.l, args.params.r or params.b')

if args.extend:
	left   = params.get('l', params.get('b', 0))
	right  = params.get('r', params.get('b', 0))
	stdns  = params.get('s', False)
	reader = TsvReader(infile, cnames = False)
	writer = TsvWriter(outfile)
	for r in reader:
		if not stdns or r[5] == '+':
			left2, right2 = left, right
		else:
			left2, right2 = right, left
		if params.pct:
			length   = r[2] - r[1]
			r[1] -= round(length * left2)
			r[2] += round(length * right2)
		else:
			r[1] -= left2
			r[2] += right2
		writer.write(r)
else:
	params._stdout = outfile
	bedtools.flank(**params).run()




from bioprocs.utils import runcmd, cmdargs
from bioprocs.utils.tsvio import TsvReader, TsvWriter, TsvRecord

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
liftover = {{args.liftover | quote}}
lochain  = {{args.lochain | quote}}
genome   = {{args.genome | quote}} # target genome

bedfile      = "{{job.outdir}}/{{i.infile | fn}}.bed"
mappedfile   = "{{job.outdir}}/{{i.infile | fn}}.mapped.bed"
unmappedfile = "{{job.outdir}}/{{i.infile | fn}}.unmapped.bed"

# get a bed file from maf file
reader = TsvReader(infile, ftype = 'head')
writer = TsvWriter(bedfile, ftype = 'bed')

for i, r in enumerate(reader):
	bedr        = TsvRecord()
	bedr.CHR    = r.Chromosome
	bedr.START  = r.Start_Position
	bedr.NAME   = 'MAF{}'.format(i)
	bedr.SCORE  = int(r.End_Position) - int(r.Start_Position)
	bedr.STRAND = r.Strand
	# liftOver doesn't work with bed index start with 0
	bedr.END = int(r.End_Position) + 1 if bedr.SCORE == 0 else r.End_Position
	writer.write(bedr)

writer.close()
reader.rewind()

# run liftOver
cmd = '{liftover} {bedfile} {chainfile} {mappedfile} {unmappedfile}'
runcmd(cmd.format(
	liftover     = liftover,
	bedfile      = bedfile,
	chainfile    = lochain,
	mappedfile   = mappedfile,
	unmappedfile = unmappedfile
))

# read all mapped file
reader2 = TsvReader(mappedfile, ftype = 'bed')
mapped = {}
for r in reader2:
	if r.SCORE == 0:
		mapped[r.NAME] = r.CHR, r.START, r.START
	else:
		mapped[r.NAME] = r.CHR, r.START, r.END

# mapped them back
writer2 = TsvWriter(outfile)
writer2.meta = reader.meta
writer2.writeHead()
for i, r in enumerate(reader):
	name = 'MAF{}'.format(i)
	if not name in mapped: continue
	Chr, Start, End = mapped[name]
	r.Chromosome, r.Start_Position, r.End_Position = mapped[name]
	r.NCBI_Build = genome
	writer2.write(r)
		



import gzip
from os import path, environ
from cyvcf2 import VCF, Writer
from pyppl import Box
from difflib import SequenceMatcher
from bioprocs.utils import shell2 as shell

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
errfile = {{job.errfile | quote}}
# currently supported:
# clinvarLink  = True,
# addChr       = True,
# headerInfo   = {id1: {Number:1, Type:String, Description: xxx}},
# headerContig = {ref: xxx.fai/xxx.dict, notfound: drop}
# headerFormat = {id1: {Number:1, Type:String, Description: xxx}},
# headerFilter = {filter1: descirption1, filter2: desc2, ...}

fixes   = {{args.fixes | repr}}
nthread = {{args.nthread | repr}}
ref     = {{args.ref | quote}}
reffai  = ref + '.fai'
refdict = ref[:-3] + '.dict'

environ['OPENBLAS_NUM_THREADS'] = str(nthread)
environ['OMP_NUM_THREADS']      = str(nthread)
environ['NUMEXPR_NUM_THREADS']  = str(nthread)
environ['MKL_NUM_THREADS']      = str(nthread)
import numpy

inverse            = fixes.get('_inverse', False)
fixes.clinvarLink  = fixes.get('clinvarLink', False)
fixes.addChr       = fixes.get('addChr', False)
fixes.addAF        = fixes.get('addAF', False)
fixes.tumorpos     = fixes.get('tumorpos', False)
fixes.headerInfo   = fixes.get('headerInfo', False)
fixes.headerContig = fixes.get('headerContig', False)
fixes.headerFormat = fixes.get('headerFormat', False)
fixes.headerFilter = fixes.get('headerFilter', False)
fixes.non_ref      = fixes.get('non_ref', False)

if inverse:
	inv = lambda v: not v if isinstance(v, bool) else v
	fixes.clinvarLink  = inv(fixes.clinvarLink)
	fixes.addChr       = inv(fixes.addChr)
	fixes.addAF        = inv(fixes.addAF)
	fixes.tumorpos     = inv(fixes.tumorpos)
	fixes.headerInfo   = inv(fixes.headerInfo)
	fixes.headerContig = inv(fixes.headerContig)
	fixes.headerFormat = inv(fixes.headerFormat)
	fixes.headerFilter = inv(fixes.headerFilter)
	fixes.non_ref      = inv(fixes.non_ref)

if fixes.tumorpos is True:
	fixes.tumorpos = path.splitext(path.basename(infile))[0]
if fixes.tumorpos and isinstance(fixes.tumorpos, str):
	fixes.tumorpos = fixes.tumorpos.split(',')

def getContigsFromFai(fai):
	ret = {}
	with open(fai) as f:
		for line in f:
			parts = line.split('\t')
			ret[parts[0]] = int(parts[1])
	return ret

def getContigsFromDict(dictfile):
	ret = {}
	with open(dictfile) as f:
		for line in f:
			parts = line.split('\t')
			if parts[0] != '@SQ':
				continue
			ret[parts[1][3:]] = int(parts[2][3:])
	return ret

# fix clinvarLink first, because it will fail VCF parser
# just remove it
if fixes.clinvarLink or fixes.addChr or fixes.tumorpos or fixes.non_ref:
	tmpoutfile = outfile + '.tmp'
	openfun = gzip.open if infile.endswith('.gz') else open
	with openfun(infile, 'rt', errors='replace') as fin, \
		openfun(tmpoutfile, 'wt', errors='replace') as fout:
		for line in fin:
			if line.startswith('##'):
				# if no header operated
				fout.write(line)
			elif line.startswith('#') and fixes.tumorpos:
				# determine if we have 2 samples
				parts = line.strip().split('\t')
				if len(parts) != 11:
					fixes.tumorpos = False
				elif parts[10] in fixes.tumorpos:
					fixes.tumorpos = True
				elif parts[9] in fixes.tumorpos:
					fixes.tumorpos = False
				else:
					sample1_match = max(SequenceMatcher(None, parts[9], tumor).ratio()
						for tumor in fixes.tumorpos)
					sample2_match = max(SequenceMatcher(None, parts[10], tumor).ratio()
						for tumor in fixes.tumorpos)
					fixes.tumorpos = sample2_match > sample1_match
				if fixes.tumorpos:
					parts[9], parts[10] = parts[10], parts[9]
				fout.write('\t'.join(parts) + '\n')
			elif line.startswith('#'):
				fout.write(line)
			else:
				parts = line.strip().split('\t')
				if fixes.tumorpos:
					parts[9], parts[10] = parts[10], parts[9]
				if fixes.addChr:
					parts[0] = parts[0] if parts[0].startswith('chr') else 'chr' + parts[0]
				if fixes.non_ref and "<NON_REF>" in parts[4]:
					parts[4] = ','.join(alt for alt in parts[4].split(',') if alt != '<NON_REF>') \
						or '.'
				info = parts[7]
				if fixes.clinvarLink:
					info = ';'.join(inf for inf in info.split(';') if not inf.startswith('<a href'))
				parts[7] = info

				fout.write('\t'.join(parts) + '\n')
	infile = tmpoutfile

pool = {}
if fixes.headerInfo not in (None, False):
	pool['info'] = set()
if fixes.headerContig not in (None, False):
	pool['contig'] = set()
if fixes.headerFormat not in (None, False):
	pool['format'] = set()
if fixes.headerFilter not in (None, False):
	pool['filter'] = set()

if pool or fixes.addAF:
	# scan the problems
	vcf  = VCF(infile)

	for variant in vcf:
		if fixes.headerContig not in (None, False):
			pool['contig'].add(variant.CHROM)
		if fixes.headerInfo not in (None, False):
			for key,_ in variant.INFO:
				pool['info'].add(key)
		if fixes.headerFormat not in (None, False):
			for key in variant.FORMAT:
				pool['format'].add(key)
		if fixes.headerFilter not in (None, False) and variant.FILTER:
			for filt in variant.FILTER.split(';'):
				pool['filter'].add(filt)
	vcf.close() # remove auto-generated headers

	if fixes.headerContig is True:
		fixes.headerContig = {}
	if fixes.headerInfo is True:
		fixes.headerInfo = {}
	if fixes.headerFormat is True:
		fixes.headerFormat = {}
	if fixes.headerFilter is True:
		fixes.headerFilter = {}

	vcf  = VCF(infile)

	# if fixes.addAF is true, try to add it to header
	if fixes.addAF:
		try:
			vcf["AD"]
		except KeyError:
			raise ValueError('Cannot fix addAF, vcf file does not contain "FORMAT/AD" field.')
		try:
			vcf["DP"]
		except KeyError:
			raise ValueError('Cannot fix addAF, Vcf file does not contain "FORMAT/DP" field.')
		try:
			item_type = vcf["AF"]
			if item_type['Description'] == '"Dummy"':
				raise KeyError()
		except KeyError:
			adict = dict(
				ID          = "AF",
				Number      = 1,
				Type        = "Float",
				Description = "The allele frequency of first minor allele."
			)
			vcf.add_format_to_header(adict)

	for info_item in pool.get('info', []):
		try:
			item_type = vcf[info_item]
			# auto added by cyvcf2
			if item_type['Description'] == '"Dummy"':
				raise KeyError()
		except KeyError:
			adict = fixes.headerInfo.get(info_item, {})
			adict = dict(
				ID          = info_item,
				Number      = adict.get('Number', 1),
				Type        = adict.get('Type', "String"),
				Description = adict.get('Description', info_item.upper())
			)
			vcf.add_info_to_header(adict)

	for fmt_item in pool.get('format', []):
		try:
			item_type = vcf[fmt_item]
			if item_type['Description'] == '"Dummy"':
				raise KeyError()
		except KeyError:
			adict = fixes.headerFormat.get(fmt_item, {})
			adict = dict(
				ID          = fmt_item,
				Number      = adict.get('Number', 1),
				Type        = adict.get('Type', "String"),
				Description = adict.get('Description', fmt_item.upper())
			)
			vcf.add_format_to_header(adict)

	for filter_item in pool.get('filter', []):
		try:
			item_type = vcf[filter_item]
		except KeyError:
			desc  = fixes.headerFilter.get(filter_item, filter_item.upper())
			adict = dict(
				ID          = filter_item,
				Description = desc
			)
			vcf.add_filter_to_header(adict)

	contigs2drop = set()
	if pool.get('contig'):
		if path.isfile(reffai):
			refcontigs = getContigsFromFai(reffai)
		elif path.isfile(refdict):
			refcontigs = getContigsFromDict(refdict)
		else:
			refcontigs = {contig_item: 999999999 for contig_item in pool['contig']}

		for contig_item in pool['contig']:
			if contig_item not in vcf.seqnames:
				contig_len = refcontigs.get(contig_item, 999999999)
				vcf.add_to_header("##contig=<ID=%s,length=%s>" % (contig_item, contig_len))

	writer = Writer(outfile, vcf)
	#vcf.close()

	for variant in vcf:
		if variant.CHROM in contigs2drop:
			continue
		if fixes.addAF:
			ADs = variant.format("AD")
			if ADs is None:
				continue
			DPs = variant.format("DP")
			variant.set_format("AF", numpy.array([
				numpy.float(ad[1])/numpy.float(DPs[i][0])
				if DPs[i] and float(DPs[i][0]) > 0 else 0.0
				for i, ad in enumerate(ADs)]))
		writer.write_record(variant)
	writer.close()

else:
	shell.mv(infile, outfile)

with open(errfile, 'r') as f:
	for line in f:
		if line.startswith('[E::vcf_parse_format]'):
			raise ValueError(line)
